"""
Test vectors for internal use and to assist in cross verification of REUNION
protocol implementations.

>>> len(esk_a_seed) == 32
True
>>> len(esk_a) == 32
True
>>> len(epk_a) == 32
True
>>> len(pk_a) == 32
True
>>> len(esk_b_seed) == 32
True
>>> len(esk_b) == 32
True
>>> len(epk_b) == 32
True
>>> len(pk_b) == 32
True
>>> len(aead_key) == 32
True
>>> len(aead_pt) == 25
True
>>> len(aead_ad) == 32
True
>>> len(aead_ct) == 41
True
>>> len(prp_key) == 32
True
>>> len(prp_pt) == 32
True
>>> len(prp_ct) == 32
True
>>> len(a1) == 32
True
>>> len(a2) == 32
True
>>> len(pdk1) == 32
True
>>> len(pdk2) == 32
True
>>> len(h) == 32
True
>>> len(ReunionSession_A_msg) == 47
True
>>> len(ReunionSession_B_msg) == 439
True
>>> len(ReunionSession_passphrase) == 10
True
>>> len(ReunionSession_four_party_A_msg) == 9
True
>>> len(ReunionSession_four_party_B_msg) == 9
True
>>> len(ReunionSession_four_party_C_msg) == 9
True
>>> len(ReunionSession_four_party_D_msg) == 9
True
"""

from rendez.vous.reunion.constants import DEFAULT_ARGON_SALT, DEFAULT_HKDF_SALT

argon2i_salt: bytes = DEFAULT_ARGON_SALT
argon2i_password: bytes = bytes(b'REUNION is for rendezvous')
argon2i_hash: bytes = bytes.fromhex('131f782cae57faa5055277621aec7c3984fbef048c8d183848f3def2697c7acd')

hkdf_salt: bytes = DEFAULT_HKDF_SALT
hkdf_key = bytes.fromhex('513e3c670ab00a436de0d801b07e085149ef205d27807d656253cd9a08a7bdf0')
hkdf_pdk = bytes.fromhex('9a3b6d37987a9ea05709a9ef2b8c8e4e0b0c51088cb6edc93bcacf4ff36fda1c')

esk_a_seed: bytes = bytes.fromhex('e60498784e625a21d6285ee7a6144a0464dab10120b11f3794dd00e36da98c27')
esk_a: bytes = bytes.fromhex('f988f98f466ff8585598ad12956b385e6090e9fdfdac3ca17c77cad61ac8a430')
epk_a: bytes = bytes.fromhex('b92b89f7bea9d4deee61a07a930edc4f50a7e5eb38a6b5667f44dea5032703f5')
pk_a: bytes = bytes.fromhex('95fa3b2a70e42f4dc66117a02680ddfe45a55451654e7bd685ba2a4179289104')
esk_b_seed: bytes = bytes.fromhex('f50a1248b83f07c6232485508bc889352531a5387b18580d8f6685c352c454d2')
esk_b: bytes = bytes.fromhex('8ba80391df517ee3e3901046adf8c4aab8068cb9a569349e98ee8241b7fde770')
epk_b: bytes = bytes.fromhex('9c1c114b9f11908e6f046805c97a1ba8261e3a3a34cfca9a72d20f3701c553b1')
pk_b: bytes = bytes.fromhex('6d4d5132efddd1ccfdb42178d5cab993617b50a43e24a0b6679e0d6f17ddae1e')
aead_key: bytes = bytes.fromhex('2e845d6aa49d50fd388c9c7072aac817ec71e323a4d32532263a757c98404c8a')
aead_pt: bytes = bytes.fromhex('5245554e494f4e20697320666f722052656e64657a766f7573')
aead_ad: bytes = bytes.fromhex('e7bab55e065f23a4cb74ce9e6c02aed0c31c90cce16b3d6ec7c98a3ed65327cf')
aead_ct: bytes = bytes.fromhex('a405c2d42d576140108a84a08a9c8ee140d5c72c5332ec6713cf7c6fb27719a9007606f7834853245b')
prp_key: bytes = bytes.fromhex('37620a87ccc74b5e425164371603bd96c794594b7d07e4887bae6c7f08fa9659')
prp_pt: bytes = bytes.fromhex('5245554e494f4e20697320666f722052656e64657a766f75732e2e2e20505250')
prp_ct: bytes = bytes.fromhex('a74b26c607e56b1f59a84d91ff738e6b55f94ceedc418118347c2b733e5ebe92')
a1: bytes = bytes.fromhex('fbe519150e9cb72815951bb49fee855c1ba3f1b8b6cdcb48013141eeb52203ba')
a2: bytes = bytes.fromhex('991f924198039449b27f61490d3a75ecf2a57795179801a40f61953453b748c9')
pdk1: bytes = bytes.fromhex('2938568958db545bf6a9a9f4b6b0f5567f1b7d45c5357c7221f80bd9dec011f3')
pdk2: bytes = bytes.fromhex('3e237c4afe43755a9a932e02233470ef4f44877341709837ae3acf680c1a301a')
h: bytes = bytes.fromhex('1ffb4f05cb3e841d44079afbcc51f62edbd7092294edac59846b8519f48c5a45')
h_preimage: bytes = bytes(b'REUNION is for rendezvous')

x25519_sk_seed_a: bytes = bytes.fromhex('a0f5f44533e439e9aced82d38eaab109df03c6f26833530343b1fac080fc6287')
x25519_sk_seed_b: bytes = bytes.fromhex('31a09e46971b29b5a9c59706c973d4f7f00361b442fd08b4724103b0b7f3ab24')
highctidh_drng_seed: bytes = bytes.fromhex('163d228fd8182bdb0e259fbf0ed5a776b47126ba4d61d774cce87f6546f8d677')
highctidh_context: int = 1

#  generate_hidden_key_pair
hidden_key_pair_seed: bytes = bytes.fromhex('5aace7eec7f3a5ead537d23cbee29ed1003f3aa73d9a7a97b72d249b9119d409')
hidden_key_pair_pk_a: bytes = bytes.fromhex('dd134b5b287d6698f8db9cd58f7f4ccd2293103010fd2e7a11ed984debe2cde6')
hidden_key_pair_sk_a: bytes = bytes.fromhex('d6b067b9b98e9616dde7e9aa52bd75f13493897ec4908230508b5abb293a5140')

# generate_ctidh_key_pair
ctidh_key_pair_seed: bytes = bytes.fromhex('4141414141414141414141414141414141414141414141414141414141414141')
ctidh_key_pair_seed_pk: bytes = bytes.fromhex('a0e897b81374cc17aa917637cda97a56377c9b7bdbe86a53a6f01ce35a0366684568e7de4e38000214a2600ac6a9d07b2379ccccdf0c7ca94ff1288eeb06347101be8cabd24543315eb1d00596d05ebfcde4f13e076bc30635db8aa249b55c992ecb24f9ba128a90b8b1d93420ca8f6454572d4c3b492027b942fb45d1e5a20e')
ctidh_key_pair_seed_sk: bytes = bytes.fromhex('01fffd00ff000000ff03ff00fd00ff00fe00000000ffff0100ffff01ff0200ff0100ffff01010001fffffe0001020001010000ff03000100ff00ff0000fd0000fe0003010100ff0302000000ff000000fe000002010001ffff00000000fe03000001ff0001fe010000010000ff00ff0100ffff00010101000000000000000100ff00')

t1_empty_id: bytes = bytes.fromhex('786a02f742015903c6c6fd852552d272912f4740e15847618a86e217f71f5419')

ReunionSession_A_msg: bytes = "Mr. Watson — Come here — I want to see you.".encode()
ReunionSession_B_msg: bytes = """\
when a man gives his order to produce a definite result and stands by that
order it seems to have the effect of giving him what might be termed a second
sight which enables him to see right through ordinary problems. What this power
is I cannot say; all I know is that it exists and it becomes available only
when a man is in that state of mind in which he knows exactly what he wants and
is fully determined not to quit until he finds it.""".encode()
ReunionSession_passphrase: bytes = b"passphrase"
ReunionSession_passphrase1: bytes = b"passphrase1"
ReunionSession_passphrase2: bytes = b"passphrase2"
ReunionSession_four_party_A_msg: bytes = b"a message"
ReunionSession_four_party_B_msg: bytes = b"b message"
ReunionSession_four_party_C_msg: bytes = b"c message"
ReunionSession_four_party_D_msg: bytes = b"d message"

# Deterministic Session vectors
dsession_salt: bytes = DEFAULT_HKDF_SALT
dsession_seed: bytes = bytes.fromhex('0123456789')
dsession_seed_a: bytes = b'alice' + dsession_seed
dsession_seed_b: bytes = b'bob' + dsession_seed
dsession_passphrase: bytes = b'reunion is for rendezvous'
dsession_dh_seed_a: bytes = bytes.fromhex('324d226178bc2e0f625dcb91d83cb7fd8ed710755695559927fc75edff85a96b') # bytes(Hash(dsession_seed_a + b"diffie-hellman")).hex()
dsession_dh_seed_b: bytes = bytes.fromhex('79455c612276de0d2d511744936efc30c4cbb0e9715b737dbaca8d7d4acaf8b0') # bytes(Hash(dsession_seed_b + b"diffie-hellman")).hex()

# To aid developers who do not have a highctidh seed keygen interface we
# include the following vectors and document their generation
# from rendez.vous.reunion.__vectors__ import dsession_seed # bytes.fromhex('0123456789')
# from rendez.vous.reunion.primitives import Hash
# from rendez.vous.reunion.primitives import generate_ctidh_key_pair
# ctidh_seed=bytes(Hash(dsession_seed + b'ctidh'))
# pk, sk = generate_ctidh_key_pair(seed=ctidh_seed)
# pk_bytes = bytes(pk)
# sk_bytes = bytes(sk)
# pk_bytes.hex()
# '4bcac087636d43ba832e08b95fd2a3f39f399baa9b6104f7b77ce224d1542f1480e6129a77e321da2c140da6fc084fcf16c57223a47913be2a3bddba78f09a012cfbe63cd8e7dd1f6253851067a17d725407ce12b8c115e977bdee9035c4981e0eca26df65f6b27cd0dd5bc64b765d890ecf9c486148243851bbf92d1df7e80d'
# sk_bytes.hex()
# '0200020000fffe00fe00ff00010001ff000001ffff0100ff000202ffff0000ff01000101fe000002000200fe0000fffffffd0000ff01000000ff0001fe0100ff000101010100ffff00ff00ff000301010100ffffff010300ff00fe00000201000000010101000000ffffff010001010100010100ffff000001ff00000000ff000100'
dsession_ctidh_seed_a: bytes =  bytes.fromhex('72122f0b686abcbe62a15e2bc819b97e30f66da2baec7c9fb4372ff74e7883e9') # bytes(Hash(dsession_seed_a + b"ctidh")).hex()
dsession_ctidh_seed_b: bytes =  bytes.fromhex('be3e954aa84dc8526dce9774f099175f9bc86e0cca0eb357d5b79484678e21c6') # bytes(Hash(dsession_seed_b + b"ctidh")).hex()

dsession_ctidh_seed_sk_a: bytes = bytes.fromhex('0200fd0000ff01fffe00fe000100fe00000001fefe00010000fd0001000000fd0000fe00ff0200fffdff010000ff000201ff00fe0100ff000000010000ff030000010100020101ff0002ff00ff0001000002000002ff0101ff010000ff01020100000000fe0000ff000100ff00ff000100ff000000ff000001ff02ff010000000100') # _, sk = generate_ctidh_key_pair(seed=dsession_ctidh_seed_a); bytes(sk).hex()
dsession_ctidh_seed_pk_a: bytes = bytes.fromhex('45a65c2972ddf00acf524db29bf9394ce79a17fdb08f1f135dd8b3296d4d83281381b039312a8fbb19a982bea2f55b71e7998c125aeea20eebbbd683b556f35bf250c0b07c7b59251b36610110aa506ddeb8400df688f2560fc9ab8c830786acd47a6e356ae9f9bbece3df3c4b6e083bde2a955fe583e79071c05e8348280104') # pk, _ = generate_ctidh_key_pair(seed=dsession_ctidh_seed_a); bytes(pk).hex()

dsession_ctidh_seed_sk_b: bytes = bytes.fromhex('01ff00fffd0100000004ff040000000002020100ff0002ff000000020001fd0000ff00ff010301000103000000000000fe03ff0000ffffff02010000ff00fd01010002ff00fe010100ff01ffffff00ff0100010100030100000000000100fc00fe00000001fe0000010000ff000200010100000000000200fe01ff00ff0001000000') # _, sk = generate_ctidh_key_pair(seed=dsession_ctidh_seed_b); bytes(sk).hex()
dsession_ctidh_seed_pk_b: bytes = bytes.fromhex('abfea79fc55ef3557b3e39fe01613b2d789e01d26a018d02e2388603a734110e478699d7e90292a393909009975889be1719aff29edbf6a3ec170589689840cf615fff22aa1b37abb9d9ba010953ee154c78f9e0eee28d625a7d2df73b5aa22b502b5040edb7fe8feb48b13541fb647974a3e5d24441e9021da8ad37fefeca05') # pk, _ = generate_ctidh_key_pair(seed=dsession_ctidh_seed_b); bytes(pk).hex()

dsession_gamma_seed_a: bytes = bytes.fromhex('6ba51aada3aca321534d73733860b59ea63a9746dc0bd3b00c09f5eb6feb508a') # bytes(Hash(dsession_seed_a + b"gamma")).hex()
dsession_gamma_seed_b: bytes = bytes.fromhex('75884ac7ad53827bb7bf280bc016191bcdfb6ef80c434e8155ef102e8db258ce') # bytes(Hash(dsession_seed_b + b"gamma")).hex()

dsession_delta_seed_a: bytes = bytes.fromhex('d5d0587357083f14ba559f775432b948f30e8e658ff866e2873b7768b3fa8ba5') # bytes(Hash(dsession_seed_a + b"delta")).hex()
dsession_delta_seed_b: bytes = bytes.fromhex('33f1732459211686a4acf28f0cccaa0b8cb9f57b5398765481cd073297a38449') # bytes(Hash(dsession_seed_b + b"delta")).hex()

dsession_dummy_seed_a: bytes = bytes.fromhex('a99f86fb345e9d833ce5534df39beb076f48c4cb62cdb940e23324df510065ea') # bytes(Hash(dsession_seed_a + b"dummy")).hex()
dsession_dummy_seed_b: bytes = bytes.fromhex('9e7a7f9f95146604a206a1a577f6d34dc9550054ae635d955eef9b33a8a899b9') # bytes(Hash(dsession_seed_b + b"dummy")).hex()

dsession_tweak_a: int = 32 # Hash(dsession_seed_a + b"tweaked")[0]
dsession_tweak_b: int = 243 # Hash(dsession_seed_b + b"tweaked")[0]

dsession_payload_A: bytes = bytes.fromhex('f06b1a5db24a0394fb28a53de02059fc34166424e40e64d7a857efdc38f158f1') # bytes(Hash(b"deterministic session payload A" + dsession_seed_a)).hex()
dsession_payload_B: bytes = bytes.fromhex('03475cb34f16bceabe4945197cf2a0064eb3a28601fc9489e613debe5e282e1d') # bytes(Hash(b"deterministic session payload B" + dsession_seed_b)).hex()
# These outputs are the known values for the generated Protocol messages
dsession_AT1: bytes = bytes.fromhex('e668c52c59cacc162dc6e36dcf42b6ee861a5765b45d8f83c25447c25fb3b49245a65c2972ddf00acf524db29bf9394ce79a17fdb08f1f135dd8b3296d4d83281381b039312a8fbb19a982bea2f55b71e7998c125aeea20eebbbd683b556f35bf250c0b07c7b59251b36610110aa506ddeb8400df688f2560fc9ab8c830786acd47a6e356ae9f9bbece3df3c4b6e083bde2a955fe583e79071c05e834828010496515ab6b675874566529d859a131b50b05bebaa197fe1920b42d1eff6fecc4aac538ab23f58f17b80d131cdb5e112c53c3bd6e4f0cf2df35f1e7d668b2ad1df')
dsession_AT1_alpha: bytes = bytes.fromhex('e668c52c59cacc162dc6e36dcf42b6ee861a5765b45d8f83c25447c25fb3b492')
dsession_AT1_beta: bytes = bytes.fromhex('45a65c2972ddf00acf524db29bf9394ce79a17fdb08f1f135dd8b3296d4d83281381b039312a8fbb19a982bea2f55b71e7998c125aeea20eebbbd683b556f35bf250c0b07c7b59251b36610110aa506ddeb8400df688f2560fc9ab8c830786acd47a6e356ae9f9bbece3df3c4b6e083bde2a955fe583e79071c05e8348280104')
dsession_AT1_gamma: bytes = bytes.fromhex('96515ab6b675874566529d859a131b50')
dsession_AT1_delta: bytes = bytes.fromhex('b05bebaa197fe1920b42d1eff6fecc4aac538ab23f58f17b80d131cdb5e112c53c3bd6e4f0cf2df35f1e7d668b2ad1df')
dsession_AT2: bytes = bytes.fromhex('f1872b8c7d23cc9702b4b118d5bb60bfc25f5c54c177fd7929c4f5af0a68eae5')
dsession_AT3: bytes = bytes.fromhex('74d2fe513e03e8fbe44385164e964f536e0c32c6af07ba2a71596a1fec33a711')
dsession_BT1: bytes = bytes.fromhex('913b88079eab1350bf1bb9f8733dc001f32cf5438448c0adaa79d20537a9964eabfea79fc55ef3557b3e39fe01613b2d789e01d26a018d02e2388603a734110e478699d7e90292a393909009975889be1719aff29edbf6a3ec170589689840cf615fff22aa1b37abb9d9ba010953ee154c78f9e0eee28d625a7d2df73b5aa22b502b5040edb7fe8feb48b13541fb647974a3e5d24441e9021da8ad37fefeca053bfeccc3e28e9e9bc334ea418a31a2bbbd2a144eb045a30f91c3bd6cb39f82c703125f7620e44935f3ed76e540ca839980f208afa40d2773eb60b35ed8ac9a26')
dsession_BT1_alpha: bytes = bytes.fromhex('913b88079eab1350bf1bb9f8733dc001f32cf5438448c0adaa79d20537a9964e')
dsession_BT1_beta: bytes = bytes.fromhex('abfea79fc55ef3557b3e39fe01613b2d789e01d26a018d02e2388603a734110e478699d7e90292a393909009975889be1719aff29edbf6a3ec170589689840cf615fff22aa1b37abb9d9ba010953ee154c78f9e0eee28d625a7d2df73b5aa22b502b5040edb7fe8feb48b13541fb647974a3e5d24441e9021da8ad37fefeca05')
dsession_BT1_gamma: bytes = bytes.fromhex('3bfeccc3e28e9e9bc334ea418a31a2bb')
dsession_BT1_delta: bytes = bytes.fromhex('bd2a144eb045a30f91c3bd6cb39f82c703125f7620e44935f3ed76e540ca839980f208afa40d2773eb60b35ed8ac9a26')
dsession_BT2: bytes = bytes.fromhex('12424018c8a15ca1d88247ef1285b3f8d36fffe33090a5af87b453acb2e7626e')
dsession_BT3: bytes = bytes.fromhex('4e01f3a88fa886b41f786dab1ccf82f94a73082bbb8444c0408b875187355b9a')



"""
This module implements the REUNION cryptographic protocol as described in
Algorithm 1 of the REUNION paper.

>>> assert len(DEFAULT_HKDF_SALT) == 32
"""

import os
from typing import Dict, List


# Curve25519 E/F p with Diﬀie-Hellman function DH(private, public), base point
# P ∈ E(F p ), 32-byte EpochID, 32-byte SharedRandom, aead-enc(key, plaintext,
# ad) function, aead-dec(key, ciphertext, ad) function, password hashing
# function argon2id(), key derivation function HKDF(), 32-byte block cipher
# rijndael-enc(key, plaintext)/rijndael-dec(key, ciphertext), random number
# generator function RNG, Elligator encode/decode functions.

from rendez.vous.reunion.primitives import (
    x25519,
    ctidh1024,
    aead_encrypt,
    aead_decrypt,
    argon2i,
    hkdf,
    prp_encrypt,
    prp_decrypt,
    unelligator,
    generate_hidden_key_pair,
    generate_ctidh_key_pair,
    Hash,
)

from rendez.vous.reunion.constants import (
    DEFAULT_HKDF_SALT
)


class T1(bytes):

    """
    This class provides attribute access to the four parts of a T1 message, and
    its id.

    The T1 message consists of four sections; the α, β, γ, and δ:
    – α is an Elligator-encoded Curve25519 public key, encrypted by a PRP using
    a symmetric key derived as described in Step 8. The encryption and
    decryption of α is unauthenticated, i.e. does not provide a validity check.
    The elligator encoding ensures that plaintext is indistinguishable from
    random bytes and every sequence of random bytes maps to a valid public key.
    – β is a CSIDH public key
    – γ is the MAC of an AEAD encryption of an empty string using a random
    key which is revealed by the T2.
    – δ is an AEAD ciphertext containing the message payload which is only de-
    cryptable by a valid T3.


    >>> from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase
    >>> from rendez.vous.reunion.__vectors__ import ReunionSession_A_msg
    >>> ReunionSession_a = ReunionSession.create(ReunionSession_passphrase, ReunionSession_A_msg)
    >>> T1 = ReunionSession_a.t1
    >>> alpha = T1.alpha
    >>> type(alpha) == bytes
    True
    >>> len(alpha) == T1.LEN_ALPHA
    True
    >>> beta = T1.beta
    >>> type(beta) == bytes
    True
    >>> len(beta) == T1.LEN_BETA
    True
    >>> gamma = T1.gamma
    >>> type(beta) == bytes
    True
    >>> len(gamma) == T1.LEN_GAMMA
    True
    """

    LEN_ALPHA = 32
    LEN_BETA = 128
    LEN_GAMMA = 16

    @property
    def alpha(self):
        """
        α is an Elligator-encoded Curve25519 public key, encrypted by a PRP using
        a symmetric key derived as described in Step 8. The encryption and
        decryption of α is unauthenticated, i.e. does not provide a validity check.
        The elligator encoding ensures that plaintext is indistinguishable from
        random bytes and every sequence of random bytes maps to a valid public key.

        Returns *T1.LEN_ALPHA* byte slice of length 32.

        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase
        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_A_msg
        >>> ReunionSession_a = ReunionSession.create(ReunionSession_passphrase, ReunionSession_A_msg)
        >>> T1 = ReunionSession_a.t1
        >>> alpha = T1.alpha
        >>> type(alpha) == bytes
        True
        >>> len(alpha) == T1.LEN_ALPHA
        True
        """
        return self[: self.LEN_ALPHA]

    @property
    def beta(self):
        """
        β is a CSIDH public key. CSIDH is provided by highctidh and CTIDH1024
        is used.

        Returns *T1.LEN_BETA* byte slice of length 128.

        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase
        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_A_msg
        >>> ReunionSession_a = ReunionSession.create(ReunionSession_passphrase, ReunionSession_A_msg)
        >>> T1 = ReunionSession_a.t1
        >>> beta = T1.beta
        >>> type(beta) == bytes
        True
        >>> len(beta) == T1.LEN_BETA
        True
        """
        return self[self.LEN_ALPHA : self.LEN_ALPHA + self.LEN_BETA]

    @property
    def gamma(self):
        """
        γ is the MAC of an AEAD encryption of an empty string using a random
        key which is revealed by the T2. AEAD is provided by monocypher
        and XChaCha20Poly1305 is used.

        Returns *T1.LEN_GAMMA* byte slice of length 16.

        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase
        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_A_msg
        >>> ReunionSession_a = ReunionSession.create(ReunionSession_passphrase, ReunionSession_A_msg)
        >>> T1 = ReunionSession_a.t1
        >>> gamma = T1.gamma
        >>> type(gamma) == bytes
        True
        >>> len(gamma) == T1.LEN_GAMMA
        True
        """
        return self[
            self.LEN_ALPHA
            + self.LEN_BETA : self.LEN_ALPHA
            + self.LEN_BETA
            + self.LEN_GAMMA
        ]

    @property
    def delta(self):
        """
        δ is an AEAD ciphertext containing the message payload which is only de-
        cryptable by a valid T3. AEAD is provided by monocypher and
        XChaCha20Poly1305 is used.

        Returns *T1.LEN_DELTA* byte slice of variable length.

        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase
        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_A_msg
        >>> ReunionSession_a = ReunionSession.create(ReunionSession_passphrase, ReunionSession_A_msg)
        >>> T1 = ReunionSession_a.t1
        >>> delta = T1.delta
        >>> type(delta) == bytes
        True
        """
        return self[self.LEN_ALPHA + self.LEN_BETA + self.LEN_GAMMA :]

    @property
    def id(self):
        """
        Returns a *Hash* of itself as 32 bytes.

        >>> from rendez.vous.reunion.__vectors__ import t1_empty_id
        >>> t1 = T1()
        >>> t1_id = t1.id
        >>> t1_empty_id.hex() == t1_id.hex()
        True
        >>> len(t1_id) == 32
        True
        """
        return Hash(self)

    def __repr__(self):
        """
        *__repr__* returns the type and six bytes of the id encoded as hex.

        >>> T1() # doctest: +ELLIPSIS
        <T1:...>
        >>> t1 = T1()
        >>> t1.__repr__() # doctest: +ELLIPSIS
        '<T1:...>'
        """
        return "<%s:%s>" % (
            type(self).__name__,
            self.id[:6].hex(),
        )


class ReunionSession(object):
    """
    Phases 0 and 1 are implemented in this object directly, while 2, 3, and 4
    are implemented in the Peer object which ReunionSession instantiates for
    each new t1.

    >>> from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase
    >>> from rendez.vous.reunion.__vectors__ import ReunionSession_A_msg
    >>> from rendez.vous.reunion.__vectors__ import ReunionSession_B_msg
    >>> ReunionSession_a = ReunionSession.create(ReunionSession_passphrase, ReunionSession_A_msg)
    >>> ReunionSession_b = ReunionSession.create(ReunionSession_passphrase, ReunionSession_B_msg)
    >>> A_t2 = ReunionSession_a.process_t1(ReunionSession_b.t1)
    >>> B_t2 = ReunionSession_b.process_t1(ReunionSession_a.t1)
    >>> A_t3, is_dummy_A = ReunionSession_a.process_t2(ReunionSession_b.t1.id, B_t2)
    >>> B_t3, is_dummy_B = ReunionSession_b.process_t2(ReunionSession_a.t1.id, A_t2)
    >>> is_dummy_A == False
    True
    >>> is_dummy_B == False
    True
    >>> B_msg_A = ReunionSession_a.process_t3(ReunionSession_b.t1.id, B_t3)
    >>> A_msg_B = ReunionSession_b.process_t3(ReunionSession_a.t1.id, A_t3)
    >>> ReunionSession_A_msg == A_msg_B
    True
    >>> ReunionSession_B_msg == B_msg_A
    True
    """

    @classmethod
    def keygen(cls):
        """
        All of the RNG access is consolidated here in a class method, so that
        everything else in the ReunionSession and Peer classes (except "create"
        which calls this) can be sans IO.

        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase
        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_A_msg
        >>> from rendez.vous.reunion.session import ReunionSession
        >>> ReunionSession_a = ReunionSession.create(ReunionSession_passphrase, ReunionSession_A_msg)
        >>> ReunionSession_a_keys = ReunionSession_a.keygen()
        >>> ReunionSession_a_keys.keys()
        dict_keys(['dh_seed', 'ctidh_seed', 'gamma_seed', 'delta_seed', 'dummy_seed', 'tweak'])
        """
        return dict(
            dh_seed=Hash(os.urandom(32)),
            ctidh_seed=Hash(os.urandom(32)),
            gamma_seed=Hash(os.urandom(32)),
            delta_seed=Hash(os.urandom(32)),
            dummy_seed=Hash(os.urandom(32)),
            tweak=Hash(os.urandom(32))[0],
        )

    @classmethod
    def create(cls, passphrase: bytes, payload: bytes, salt=DEFAULT_HKDF_SALT):
        """
        This is the typical way to instantiate a ReunionSession object.

        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase
        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_A_msg
        >>> from rendez.vous.reunion.session import ReunionSession
        >>> ReunionSession_a = ReunionSession.create(ReunionSession_passphrase, ReunionSession_A_msg)
        >>> type(ReunionSession_a)
        <class 'rendez.vous.reunion.session.ReunionSession'>
        """
        # XXX The above tests should construct a ReunionSession that sets all
        # seeds and verifies the values of the returned fields on the object.
        return cls(
            payload=payload,
            salt=salt,
            passphrase=passphrase,
            **cls.keygen(),
        )

    def __init__(
        self,
        salt: bytes,
        passphrase: bytes,
        payload: bytes,
        dh_seed: bytes,
        ctidh_seed: bytes,
        gamma_seed: bytes,
        delta_seed: bytes,
        dummy_seed: bytes,
        tweak: int,
    ):
        # dict of Peer objects, keyed by their t1 id
        self.peers: Dict = {}

        # list of payloads decrypted
        self.results: List = []

        self.dh_epk, dh_sk_bytes = generate_hidden_key_pair(dh_seed)

        # Step2a: esk Aα ∈ Z, public key epk Aα = esk Aα · P ∈ E(Fp).
        self.dh_sk: bytes = dh_sk_bytes

        # Step2b: esk Aβ ∈ Z, public key epk Aβ = esk Aβ · P ∈ E(Fp)
        self.csidh_pk, self.csidh_sk = generate_ctidh_key_pair(ctidh_seed)

        # Step 3: salt ← SharedRandom∥EpochID.
        # Setting the salt, or context, is the responsibility of the
        # application using this library. DEFAULT_SALT is 32 null bytes.
        self.salt = salt

        # Step 4: pdk ← HKDF(salt, argon2id(salt, Q)).
        kdf = hkdf(key=argon2i(passphrase, salt), salt=salt)
        self.pdk = kdf.expand(b"", 32)

        # Step 5a: sk Aγ ← H(pdk, RNG(32), msg A )
        self.sk_gamma = Hash(self.pdk + gamma_seed + payload)

        # Step 5b: sk Aδ ← H(pdk, RNG(32), msg A )
        self.sk_delta = Hash(self.pdk + delta_seed + payload)

        # t1 beta is the unencrypted csidh pk
        beta = self.csidh_pk

        # Step 6: T1Aγ ← aead-enc(sk Aγ ,“”, RS)
        gamma = aead_encrypt(self.sk_gamma, b"", salt)

        # Step 7: T1Aδ ← aead-enc(sk Aδ , msg a , RS)
        delta = aead_encrypt(self.sk_delta, payload, salt)

        # Step 8: pdkA ← H(pdk, epkAβ , T1Aγ , T1Bδ )
        self.alpha_key = Hash(self.pdk + beta + gamma + delta)

        # Step 9: T1Aα ← rijndael-enc(pdkA , epkAα)
        alpha = prp_encrypt(self.alpha_key, self.dh_epk)

        # Step 10: T1A ← T1 Aα ∥ epkAβ ∥ T1 Aγ ∥ T1 Aδ
        self.t1 = T1(alpha + beta + gamma + delta)

        # deviation from the paper - we generate dummys derministically using this Hkdf.
        self.dummy_hkdf = hkdf(key=dummy_seed, salt=salt)
        # FIXME: why dummy_hkdf? doesn't it run out after a small number of values?
        # was this just for testing purposes?

    # steps 11, 12, and 13 happen in the application using this library

    def process_t1(self, t1: bytes):
        # Step 14: for each new T1Bi do ▷ Phase 2: Process T1; transmit T2
        # implementation of steps 14 to 22 is in Peer.__init__
        """
        Process T1 message *t1* from respective peers.

        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase
        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_A_msg
        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_B_msg
        >>> ReunionSession_a = ReunionSession.create(ReunionSession_passphrase, ReunionSession_A_msg)
        >>> ReunionSession_b = ReunionSession.create(ReunionSession_passphrase, ReunionSession_B_msg)
        >>> A_t2 = ReunionSession_a.process_t1(ReunionSession_b.t1)
        >>> B_t2 = ReunionSession_b.process_t1(ReunionSession_a.t1)
        """
        if t1 == self.t1:
            assert (
                False
            ), "we don't actually expect this to happen right now but later might return None here"
            return None

        peer = self.peers.get(t1.id)
        if peer is None:
            peer = Peer(t1, self)
            self.peers[t1.id] = peer

        return peer.t2_tx

    # steps 23 and 24, transmit t1, is in the application

    def process_t2(self, t1_id: bytes, t2: bytes):
        # Step 25: for each new T2Bi do ▷ Phase 3: Process T2, transmit T3
        # implementation of steps 25 to 33 is in Peer.process_t2
        """
        Process T2 message *t2* from respective peers keyed by *t1_id*.

        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase
        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_A_msg
        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_B_msg
        >>> ReunionSession_a = ReunionSession.create(ReunionSession_passphrase, ReunionSession_A_msg)
        >>> ReunionSession_b = ReunionSession.create(ReunionSession_passphrase, ReunionSession_B_msg)
        >>> A_t2 = ReunionSession_a.process_t1(ReunionSession_b.t1)
        >>> B_t2 = ReunionSession_b.process_t1(ReunionSession_a.t1)
        >>> A_t3, is_dummy_A = ReunionSession_a.process_t2(ReunionSession_b.t1.id, B_t2)
        >>> B_t3, is_dummy_B = ReunionSession_b.process_t2(ReunionSession_a.t1.id, A_t2)
        >>> is_dummy_A == False
        True
        >>> is_dummy_B == False
        True
        """

        if t1_id in self.peers:
            return self.peers[t1_id].process_t2(t2)
        else:
            return self.dummy_hkdf.expand(t1_id + t2, 32), True

    def process_t3(self, t1_id: bytes, t3: bytes):
        # Step 36: for each new T3Bi do ▷ Phase 4: Process T3; decrypt δ
        """
        Process T3 message *t3* from respective peers keyed by *t1_id*. For
        successful rendezvous, the respective peer's message is returned as a
        bytes object or *None* is returned.

        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase
        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_A_msg
        >>> from rendez.vous.reunion.__vectors__ import ReunionSession_B_msg
        >>> ReunionSession_a = ReunionSession.create(ReunionSession_passphrase, ReunionSession_A_msg)
        >>> ReunionSession_b = ReunionSession.create(ReunionSession_passphrase, ReunionSession_B_msg)
        >>> A_t2 = ReunionSession_a.process_t1(ReunionSession_b.t1)
        >>> B_t2 = ReunionSession_b.process_t1(ReunionSession_a.t1)
        >>> A_t3, is_dummy_A = ReunionSession_a.process_t2(ReunionSession_b.t1.id, B_t2)
        >>> B_t3, is_dummy_B = ReunionSession_b.process_t2(ReunionSession_a.t1.id, A_t2)
        >>> is_dummy_A == False
        True
        >>> is_dummy_B == False
        True
        >>> B_msg_A = ReunionSession_a.process_t3(ReunionSession_b.t1.id, B_t3)
        >>> A_msg_B = ReunionSession_b.process_t3(ReunionSession_a.t1.id, A_t3)
        >>> ReunionSession_A_msg == A_msg_B
        True
        >>> ReunionSession_B_msg == B_msg_A
        True
        """
        if t1_id in self.peers:
            return self.peers[t1_id].process_t3(t3)


class Peer(object):
    def __init__(peer, t1, session):
        """
        This method implements the inside of the for loop in Phase 2 of
        Algorithm 1.
        """

        # Step 14: for each new T1Bi do ▷ Phase 2: Process T1; transmit T2

        peer.t1 = t1
        peer.session = session

        # Step 15: pdkBi ← H(pdk, T1Biβ, T1Biγ, T1Biδ)
        peer.alpha_key = Hash(session.pdk + t1.beta + t1.gamma + t1.delta)

        # Step 16: epkBiα ← unelligator(rijndael-dec(pdkBi , T1Biα )).
        peer.dh_pk: bytes = unelligator(prp_decrypt(peer.alpha_key, t1.alpha))

        # Step 17: epkBiβ ← T1Biβ
        peer.csidh_pk = ctidh1024.public_key_from_bytes(t1.beta)

        # Step 18: dh1ssi ← H(DH(eskAα , epkBiα))
        peer.dh_ss = x25519(session.dh_sk, peer.dh_pk)

        # Step 19: dh2ssi ← H(DH(eskAβ , epkBiβ)).
        peer.csidh_ss = ctidh1024.dh(
            session.csidh_sk, peer.csidh_pk
        )  # note that this can throw exceptions, see app/reunion-client.py:process_t1(T1(t1))

        # Step 20: T2kitx ← H(pdkA, pdkBi, dh1ssi, dh2ssi)
        peer.t2key_tx = Hash(
            session.alpha_key + peer.alpha_key + peer.dh_ss + peer.csidh_ss
        )

        # Step 21: T2kirx ← H(pdkBi, pdkA, dh1ssi, dh2ssi)
        peer.t2key_rx = Hash(
            peer.alpha_key + session.alpha_key + peer.dh_ss + peer.csidh_ss
        )

        # Step 22: T2Ai ← rijndael-enc(T2kitx, skAγ)
        peer.t2_tx = prp_encrypt(peer.t2key_tx, session.sk_gamma)
        peer.t2_rx = None
        peer.payload = None

        # Step 23 and 24, transmit, is implemented in an application using this library

    def process_t2(peer, t2):
        """
        This method implements the inside of the for loop in Phase 3 of
        Algorithm 1. It returns a 2-tuple of (t3, is_dummy).
        """

        # Step 25: for each new T2Bi do ▷ Phase 3: Process T2, transmit T3

        # Step 26: skBiγ ← rijndael-dec(T2kirx, T2Bi)
        sk_gamma = prp_decrypt(peer.t2key_rx, t2)

        # Step 27: if “” = aead-dec(sk B i γ , T 1 B i γ , RS) then
        aead_res = aead_decrypt(sk_gamma, peer.t1.gamma, peer.session.salt)

        if aead_res is not None:
            assert aead_res == b"", aead_res

            peer.t2_rx = t2

            # Step 28: T3kitx ← H(T2kitx, T2Ai , T2Bi).
            t3key_tx = Hash(peer.t2key_tx + peer.t2_tx + t2)

            # Step 29: T3kirx ← H(T2kirx, T2Bi, T2Ai)
            peer.t3_key_rx = Hash(peer.t2key_rx + peer.t2_rx + peer.t2_tx)

            # Step 30: T3Ai ← rijndael-enc(T3kitx, skAδ)
            return prp_encrypt(t3key_tx, peer.session.sk_delta), False

        # Step 31: else
        else:
            # Step 32: T3Ai ← H(RNG(32))
            return peer.session.dummy_hkdf.expand(peer.t1.id + t2, 32), True

    def process_t3(peer, t3: bytes):
        """
        This method implements the inside of the for loop in Phase 4 of
        Algorithm 1.
        """

        # Step 36: for each new T3Bi do ▷ Phase 4: Process T3; decrypt δ

        if peer.t2_rx is None:
            # we didn't receive a valid t2 from this peer, so, we can't decrypt
            # their t3.
            # At the cost of a more complicated API, we could implement support
            # for out-of-order delivery by storing t3s here and attempting to
            # process them again when we receive a t2 from the corresponding
            # t1. As it is now, if we receive their t3 before their t2 we will
            # never decrypt the t1 delta payload.
            return None

        # Step 37: skBiδ ← rijndael-dec(T3kirx, T3Bi).
        sk_delta = prp_decrypt(peer.t3_key_rx, t3)

        # Step 38: if msgBi ← aead-dec(skBiδ, T1Biδ, RS) then
        peer.payload = aead_decrypt(sk_delta, peer.t1.delta, peer.session.salt)
        if peer.payload is not None:
            # Step 39: add to results
            peer.session.results.append(peer.payload)
        return peer.payload

if '__main__' == __name__:
    import doctest
    doctest.testmod(verbose=True)



"""
General utility functions

>>> get_pairs is not None
True
"""

def get_pairs(items):
    """
    Helper function to get all unique pairs from a list of items.

    >>> list(get_pairs("ABCD"))
    [('A', 'B'), ('A', 'C'), ('A', 'D'), ('B', 'C'), ('B', 'D'), ('C', 'D')]
    """
    for i in range(len(items)):
        for j in range(i + 1, len(items)):
            yield items[i], items[j]



import click

from rendez.vous.reunion.__version__ import __version__
from rendez.vous.reunion.client import client
from rendez.vous.reunion.server import server
from rendez.vous.reunion.multicast import multicast

@click.version_option(__version__)
@click.group()
@click.pass_context
def cli(ctx):
  """ rendez.vous.reunion is for rendezvous """
  pass

def main():
  """
  Entrypoint for *setup.py* *reunion* console script.

  >>> import click
  >>> from rendez.vous.reunion.client import client
  >>> from rendez.vous.reunion.server import server
  >>> from rendez.vous.reunion.multicast import multicast
  """
  cli.add_command(client)
  cli.add_command(server)
  cli.add_command(multicast)
  cli()

if __name__ == '__main__':
   import doctest
   doctest.testmod(verbose=True)



"""
This is using two sockets, so that we can listen on a fixed port while using a
randomized port to send (which is then used to track the to/from linkages
between message types).

Each client has its own epoch-like schedule, which will of course never be
perfectly aligned and could in fact be wildly different. When an epoch starts,
the t1 is multicast. When a multicast t1 is received, if we haven't seen that
t1 before, we unicast both our t1 (again) and our t2 to the sender.

This flow is not ideal but it works for the moment. Setting the interval to a
small value, near 2x the cost of a csidh operation, results in it only working
when both clients are started at approximately the same moment (otherwise they
stay out-of-sync across their epochs).

In the course of using this, it has become apparent that the reveal_once mode
might need more work. Logically it should be applied across epochs, in case
Mallory arrived early and Bob doesn't arrive until epoch N+1... but if both
parties keep running the protocol then we'll get spurious alerts about 3rd
parties that aren't really there. 🤔

>>> assert UDPListener is not None
"""
import socket
import struct
import time

import ifaddr
import click

from rendez.vous.reunion.session import ReunionSession, T1
from rendez.vous.reunion.__version__ import __version__


class UDPListener(object):
    def __init__(self, bind_addr, port=0):
        self.bind_addr = bind_addr
        self.session = None
        self.old_session = None
        self.addr_map = {}
        self.old_session = None
        self.old_addr_map = {}

        # Create the TX socket
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        # Set a timeout so the socket does not block
        # indefinitely when trying to receive data.
        self.sock.settimeout(0.2)

        # Set the time-to-live for messages to 1 so they do not
        # go past the local network segment.
        ttl = struct.pack("b", 1)
        self.sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)
        self.sock.bind((bind_addr, port))

    def bind_multicast(self, multicast_group, port):
        ifaces_by_ip = {
            ip.ip: (a.index, a.name)
            for a in reversed(ifaddr.get_adapters())
            for ip in a.ips
        }

        if_idx, if_name = ifaces_by_ip.get(self.bind_addr, (socket.INADDR_ANY, "any"))

        print(
            "Using %s on iface %s (%r) for multicast to %s:%s"
            % (self.bind_addr, if_name, if_idx, multicast_group, port)
        )

        # Create the RX socket
        self.msock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.msock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        # Bind to the server address
        self.msock.bind(("", port))
        self.msock.settimeout(0.2)

        # Tell the operating system to add the socket to
        # the multicast group on all interfaces.
        group = socket.inet_aton(multicast_group)
        mreq = struct.pack("4sL", group, if_idx)
        self.msock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

    def send(self, prefix, message, dest):
        payload = prefix + message
        print("Sending", prefix)
        #        print("Sending %s byte message %r to %s:%s" % (len(payload), payload[:5], *dest))
        sent = self.sock.sendto(payload, dest)
        if sent != len(payload):
            print("Tried to send %s bytes but only sent %s!" % (len(payload), sent))

    def poll(self):
        try:
            data, addr = self.sock.recvfrom(1024)
        except socket.timeout:
            return
        return self.process_message(data, addr)

    def poll_multicast(self):
        try:
            data, addr = self.msock.recvfrom(1024)
        except socket.timeout:
            return
        if data.startswith(b"t1"):
            return self.process_message(data, addr)

    def process_message(self, data, addr):

        #        print("Receivd %s byte message %r from %s:%s" % (len(data), data[:5], *addr))
        peer = self.addr_map.get(addr)
        if not peer:
            peer = self.old_addr_map.get(addr)
            if peer and data[:2] != b"t1":
                print(
                    "!!! received %s message for %r from previous epoch"
                    % (data[:4], peer.t1)
                )

        if data.startswith(b"t1"):
            t1 = T1(data[3:])
            if t1.id in self.session.peers:
                print("ignoring replay of %r from %r" % (t1, addr))
                return
            elif t1.id == self.session.t1.id:
                #                print("ignoring our own %r from %r" % (t1, addr))
                return
            if data[2:3] == b"r":
                print("received t1r %r from %r" % (t1, addr))
            else:
                print("received t1_ %r from %r %r" % (t1, addr, data[:5]))
                self.send(b"t1r", self.session.t1, addr)
            t2 = self.session.process_t1(t1)
            self.send(b"t2", t2, addr)
            self.addr_map[addr] = self.session.peers[t1.id]

        elif data.startswith(b"t2"):

            if not peer:
                print("ignoring t2 message from unknown peer %r" % (addr,))
                return

            t2 = data[2:]
            t3, is_dummy = peer.process_t2(t2)
            if is_dummy:
                print("decryption of t2 from %r failed; sending dummy" % (addr,))
            else:
                print("successful decryption of t2 from %r; sending reveal" % (addr,))
            self.send(b"t3", t3, addr)

        elif data.startswith(b"t3"):

            if not peer:
                print("ignoring t3 message from unknown peer %r" % (addr,))
                return
            t3 = data[2:]
            print("received t3 from %r (%r)" % (peer.t1, addr))
            res = self.process_result(peer.process_t3(t3), addr, peer)
            print("t3 result:", res)

        else:
            print("%r sent us something weird: %r" % (addr, data[:50]))

    def process_result(self, payload, addr, peer):
        if payload is not None:
            print(
                "Decrypted message from %r on %r:\n   %s\n\n"
                % (peer.t1, addr, payload.decode())
            )

    def new_session(self, passphrase, message):
        print("creating new reunion session... ", flush=True, end="")
        started = time.time()
        self.old_session, self.session = self.session, ReunionSession.create(
            passphrase, message
        )
        print(
            "OK: creating {!r} took {:.1f} seconds)".format(
                self.session.t1, time.time() - started
            )
        )
        self.old_addr_map, self.addr_map = self.addr_map, {}


def run(passphrase, message, interval, multicast_group, port, reveal_once, bind_addr):

    passphrase = passphrase.encode()
    message = message.encode()

    udp = UDPListener(bind_addr, 0)
    udp.bind_multicast(multicast_group, port)

    started = None

    while True:
        if started is None or time.time() - started > interval:
            started = time.time()
            udp.new_session(passphrase, message)
            udp.send(b"t1_", udp.session.t1, (multicast_group, port))

        udp.poll()
        udp.poll_multicast()


@click.command()
@click.version_option(__version__)
@click.option(
    "--interval",
    "-I",
    default=60,
    help="Interval at which to start new sessions",
    show_default=True,
)
@click.option("--multicast-group", default="224.3.29.71", show_default=True)
@click.option("--bind-addr", default="0.0.0.0", show_default=True)
@click.option("--port", default=9005, show_default=True)
@click.option(
    "--reveal-once",
    is_flag=True,
    help="Only reveal the message to the first person with the correct passphrase",
)
@click.option("--passphrase", prompt=True, type=str, help="The passphrase")
@click.option("--message", prompt=True, type=str, help="The message")
def multicast(*a, **kw):
    """
    REUNION on an ethernet using multicast

    If you run it with no arguments, you will be prompted for a passphrase and
    message.
    """
    assert not kw[
        "reveal_once"
    ], "sorry, the --reveal-once feature has bitrotted at the moment. It should be reimplemented in the ReunionSession object."

    run(*a, **kw)

def main(**kw):
    multicast(**kw)

if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=True)



import os
import unittest
from dataclasses import dataclass, asdict
from rendez.vous.reunion.primitives import Hash
from rendez.vous.reunion.session import ReunionSession, DEFAULT_HKDF_SALT
from rendez.vous.reunion.__vectors__ import dsession_salt, dsession_passphrase
from rendez.vous.reunion.__vectors__ import dsession_seed_a, dsession_seed_b
from rendez.vous.reunion.__vectors__ import dsession_payload_A, dsession_payload_B
from rendez.vous.reunion.__vectors__ import dsession_seed_a, dsession_seed_b
from rendez.vous.reunion.__vectors__ import dsession_dh_seed_a, dsession_dh_seed_b
from rendez.vous.reunion.__vectors__ import dsession_ctidh_seed_a, dsession_ctidh_seed_b
from rendez.vous.reunion.__vectors__ import dsession_gamma_seed_a, dsession_gamma_seed_b
from rendez.vous.reunion.__vectors__ import dsession_ctidh_seed_sk_a, dsession_ctidh_seed_pk_a
from rendez.vous.reunion.__vectors__ import dsession_ctidh_seed_sk_b, dsession_ctidh_seed_pk_b
from rendez.vous.reunion.__vectors__ import dsession_delta_seed_a, dsession_delta_seed_b
from rendez.vous.reunion.__vectors__ import dsession_dummy_seed_a, dsession_dummy_seed_b
from rendez.vous.reunion.__vectors__ import dsession_tweak_a, dsession_tweak_b
from rendez.vous.reunion.__vectors__ import dsession_AT1, dsession_AT2, dsession_AT3
from rendez.vous.reunion.__vectors__ import dsession_AT1_alpha, dsession_AT1_beta
from rendez.vous.reunion.__vectors__ import dsession_AT1_gamma, dsession_AT1_delta
from rendez.vous.reunion.__vectors__ import dsession_BT1, dsession_BT2, dsession_BT3
from rendez.vous.reunion.__vectors__ import dsession_BT1_alpha, dsession_BT1_beta
from rendez.vous.reunion.__vectors__ import dsession_BT1_gamma, dsession_BT1_delta

def DeterministicSession(
    passphrase: bytes, payload: bytes, seed: bytes, salt=DEFAULT_HKDF_SALT
):
    # seed = bytes.fromhex('0123456789')
    assert dsession_salt == DEFAULT_HKDF_SALT

    assert dsession_seed_a == seed or dsession_seed_b == seed

    assert dsession_dh_seed_a == bytes(Hash(seed + b"diffie-hellman")) or \
           dsession_dh_seed_b == bytes(Hash(seed + b"diffie-hellman"))

    assert dsession_ctidh_seed_a == bytes(Hash(seed + b"ctidh")) or \
           dsession_ctidh_seed_b == bytes(Hash(seed + b"ctidh"))

    assert dsession_gamma_seed_a == bytes(Hash(seed + b"gamma")) or \
           dsession_gamma_seed_b == bytes(Hash(seed + b"gamma"))

    assert dsession_delta_seed_a == bytes(Hash(seed + b"delta")) or \
           dsession_delta_seed_b == bytes(Hash(seed + b"delta"))

    assert dsession_dummy_seed_a == bytes(Hash(seed + b"dummy")) or \
           dsession_dummy_seed_b == bytes(Hash(seed + b"dummy"))

    assert type(dsession_tweak_a) == type(Hash(seed + b"tweaked")[0]) and \
           type(dsession_tweak_b) == type(Hash(seed + b"tweaked")[0])

    assert dsession_tweak_a == int(Hash(seed + b"tweaked")[0]) or \
           dsession_tweak_b == int(Hash(seed + b"tweaked")[0])

    assert dsession_passphrase == b'reunion is for rendezvous'

    assert dsession_payload_A == bytes(Hash(b"deterministic session payload A" + seed)) or \
           dsession_payload_B == bytes(Hash(b"deterministic session payload B" + seed))

    if seed == dsession_seed_a:
      dh_seed=dsession_dh_seed_a
      ctidh_seed=dsession_ctidh_seed_a
      gamma_seed=dsession_gamma_seed_a
      delta_seed=dsession_delta_seed_a
      dummy_seed=dsession_dummy_seed_a
      tweak=dsession_tweak_a
    if seed == dsession_seed_b:
      dh_seed=dsession_dh_seed_b
      ctidh_seed=dsession_ctidh_seed_b
      gamma_seed=dsession_gamma_seed_b
      delta_seed=dsession_delta_seed_b
      dummy_seed=dsession_dummy_seed_b
      tweak=dsession_tweak_b

    return ReunionSession(
        payload=payload,
        salt=salt,
        passphrase=passphrase,
        dh_seed=dh_seed,
        ctidh_seed=ctidh_seed,
        gamma_seed=gamma_seed,
        delta_seed=delta_seed,
        dummy_seed=dummy_seed,
        tweak=tweak,
    )


@dataclass
class SessionKAT:
    """
    This is a step toward we way that we could have regeneratable data-driven
    known-answer tests.

    Ideally the actual KATs could be read from a data file rather than this
    docstring, and could also include serialized secret keys (rather than just
    their seed as it is here now) so that other implementations could load them
    from the data file and test reunion compatibility without necessarily
    needing to be completely seed-compatible when generating the keys.

    >>> SessionKAT.generate_from_seed(bytes.fromhex('0123456789'))
          seed: 0123456789
    passphrase: 7265756e696f6e20697320666f722072656e64657a766f7573
     payload_A: f06b1a5db24a0394fb28a53de02059fc34166424e40e64d7a857efdc38f158f1
     payload_B: 03475cb34f16bceabe4945197cf2a0064eb3a28601fc9489e613debe5e282e1d
           AT1: e668c52c59cacc162dc6e36dcf42b6ee861a5765b45d8f83c25447c25fb3b49245a65c2972ddf00acf524db29bf9394ce79a17fdb08f1f135dd8b3296d4d83281381b039312a8fbb19a982bea2f55b71e7998c125aeea20eebbbd683b556f35bf250c0b07c7b59251b36610110aa506ddeb8400df688f2560fc9ab8c830786acd47a6e356ae9f9bbece3df3c4b6e083bde2a955fe583e79071c05e834828010496515ab6b675874566529d859a131b50b05bebaa197fe1920b42d1eff6fecc4aac538ab23f58f17b80d131cdb5e112c53c3bd6e4f0cf2df35f1e7d668b2ad1df
     AT1_alpha: e668c52c59cacc162dc6e36dcf42b6ee861a5765b45d8f83c25447c25fb3b492
      AT1_beta: 45a65c2972ddf00acf524db29bf9394ce79a17fdb08f1f135dd8b3296d4d83281381b039312a8fbb19a982bea2f55b71e7998c125aeea20eebbbd683b556f35bf250c0b07c7b59251b36610110aa506ddeb8400df688f2560fc9ab8c830786acd47a6e356ae9f9bbece3df3c4b6e083bde2a955fe583e79071c05e8348280104
     AT1_gamma: 96515ab6b675874566529d859a131b50
     AT1_delta: b05bebaa197fe1920b42d1eff6fecc4aac538ab23f58f17b80d131cdb5e112c53c3bd6e4f0cf2df35f1e7d668b2ad1df
           AT2: f1872b8c7d23cc9702b4b118d5bb60bfc25f5c54c177fd7929c4f5af0a68eae5
           AT3: 74d2fe513e03e8fbe44385164e964f536e0c32c6af07ba2a71596a1fec33a711
           BT1: 913b88079eab1350bf1bb9f8733dc001f32cf5438448c0adaa79d20537a9964eabfea79fc55ef3557b3e39fe01613b2d789e01d26a018d02e2388603a734110e478699d7e90292a393909009975889be1719aff29edbf6a3ec170589689840cf615fff22aa1b37abb9d9ba010953ee154c78f9e0eee28d625a7d2df73b5aa22b502b5040edb7fe8feb48b13541fb647974a3e5d24441e9021da8ad37fefeca053bfeccc3e28e9e9bc334ea418a31a2bbbd2a144eb045a30f91c3bd6cb39f82c703125f7620e44935f3ed76e540ca839980f208afa40d2773eb60b35ed8ac9a26
     BT1_alpha: 913b88079eab1350bf1bb9f8733dc001f32cf5438448c0adaa79d20537a9964e
      BT1_beta: abfea79fc55ef3557b3e39fe01613b2d789e01d26a018d02e2388603a734110e478699d7e90292a393909009975889be1719aff29edbf6a3ec170589689840cf615fff22aa1b37abb9d9ba010953ee154c78f9e0eee28d625a7d2df73b5aa22b502b5040edb7fe8feb48b13541fb647974a3e5d24441e9021da8ad37fefeca05
     BT1_gamma: 3bfeccc3e28e9e9bc334ea418a31a2bb
     BT1_delta: bd2a144eb045a30f91c3bd6cb39f82c703125f7620e44935f3ed76e540ca839980f208afa40d2773eb60b35ed8ac9a26
           BT2: 12424018c8a15ca1d88247ef1285b3f8d36fffe33090a5af87b453acb2e7626e
           BT3: 4e01f3a88fa886b41f786dab1ccf82f94a73082bbb8444c0408b875187355b9a
    """

    seed: bytes = None
    passphrase: bytes = None
    payload_A: bytes = None
    payload_B: bytes = None
    AT1: bytes = None
    AT1_alpha: bytes = None
    AT1_beta: bytes = None
    AT1_gamma: bytes = None
    AT1_delta: bytes = None
    AT2: bytes = None
    AT3: bytes = None
    BT1: bytes = None
    BT1_alpha: bytes = None
    BT1_beta: bytes = None
    BT1_gamma: bytes = None
    BT1_delta: bytes = None
    BT2: bytes = None
    BT3: bytes = None

    @classmethod
    def generate_from_seed(cls, seed):
        self = cls()
        self.seed = seed
        self.passphrase = dsession_passphrase
        self.payload_A = dsession_payload_A
        self.payload_B = dsession_payload_B
        A = DeterministicSession(self.passphrase, self.payload_A, dsession_seed_a)
        B = DeterministicSession(self.passphrase, self.payload_B, dsession_seed_b)
        self.AT1 = A.t1
        self.AT1_alpha = A.t1.alpha
        self.AT1_beta = A.t1.beta
        self.AT1_gamma = A.t1.gamma
        self.AT1_delta = A.t1.delta
        self.BT1 = B.t1
        self.BT1_alpha = B.t1.alpha
        self.BT1_beta = B.t1.beta
        self.BT1_gamma = B.t1.gamma
        self.BT1_delta = B.t1.delta
        self.AT2 = A.process_t1(B.t1)
        self.BT2 = B.process_t1(A.t1)
        self.AT3, a_isdummy = A.process_t2(B.t1.id, self.BT2)
        self.BT3, b_isdummy = B.process_t2(A.t1.id, self.AT2)
        assert not a_isdummy and not b_isdummy
        A.process_t3(B.t1.id, self.BT3)
        B.process_t3(A.t1.id, self.AT3)
        assert A.results[0] == self.payload_B
        assert B.results[0] == self.payload_A
        assert self.passphrase == dsession_passphrase
        assert self.payload_A == dsession_payload_A
        assert self.payload_B == dsession_payload_B
        assert self.AT1.hex() == dsession_AT1.hex()
        assert self.AT1_alpha.hex() == dsession_AT1_alpha.hex()
        assert self.AT1_beta.hex() == dsession_AT1_beta.hex()
        assert self.AT1_gamma.hex() == dsession_AT1_gamma.hex()
        assert self.AT1_delta.hex() == dsession_AT1_delta.hex()
        assert self.AT2.hex() == dsession_AT2.hex()
        assert self.AT3.hex() == dsession_AT3.hex()
        assert self.BT1.hex() == dsession_BT1.hex()
        assert self.BT1_alpha.hex() == dsession_BT1_alpha.hex()
        assert self.BT1_beta.hex() == dsession_BT1_beta.hex()
        assert self.BT1_gamma.hex() == dsession_BT1_gamma.hex()
        assert self.BT1_delta.hex() == dsession_BT1_delta.hex()
        assert self.BT2.hex() == dsession_BT2.hex()
        assert self.BT3.hex() == dsession_BT3.hex()
        return self

    def __repr__(self):
        return "\n".join(f"{k:>10}: {bytes(v).hex()}" for k, v in asdict(self).items())



"""
Constants used in the REUNION protocol

>>> len(DEFAULT_AEAD_NONCE) == 24
True
>>> len(DEFAULT_ARGON_SALT) == 32
True
>>> len(DEFAULT_HKDF_SALT) == 32
True
>>> DEFAULT_CTIDH_SIZE == 1024
True
"""

DEFAULT_AEAD_NONCE: bytes = b"\x00" * 24
DEFAULT_ARGON_SALT: bytes = b"\x00" * 32
DEFAULT_HKDF_SALT: bytes = b"\x00" * 32

DEFAULT_CTIDH_SIZE: int = 1024 



import os
import unittest

from rendez.vous.reunion.primitives import aead_decrypt, aead_encrypt, argon2i
from rendez.vous.reunion.primitives import generate_hidden_key_pair, Hash
from rendez.vous.reunion.primitives import highctidh_deterministic_rng, hkdf, prp_decrypt
from rendez.vous.reunion.primitives import prp_encrypt, unelligator, x25519

class TestPrimitives(unittest.TestCase):
    def test_aead(self):
        key = os.urandom(32)
        msg = os.urandom(32)
        ad = os.urandom(32)
        self.assertEqual(aead_decrypt(key, aead_encrypt(key, msg, ad), ad), msg)

    def test_elligator(self):
        epk_a, sk_a = generate_hidden_key_pair(os.urandom(32))
        epk_b, sk_b = generate_hidden_key_pair(os.urandom(32))
        ss1 = x25519(sk_a, unelligator(epk_b))
        ss2 = x25519(sk_b, unelligator(epk_a))
        assert ss1 == ss2

    def test_prp(self):
        key = os.urandom(32)
        msg = os.urandom(32)
        ct = prp_encrypt(key, msg)
        msg2 = prp_decrypt(key, ct)
        assert msg == msg2

    def test_argon_hkdf(self):
        salt = b"\x00" * 32
        passphrase = b"passphrase"
        a1 = argon2i(passphrase, salt)
        a2 = argon2i(passphrase, salt)
        assert a1 == a2
        pdk1 = hkdf(argon2i(passphrase, salt), salt).expand(b"", 32)
        pdk2 = hkdf(argon2i(passphrase, salt), salt).expand(b"", 32)
        assert pdk1 == pdk2

    def _test_argon_single_char(self):
        salt = b"\x00" * 32
        passphrase = b"p"
        a1 = argon2i(passphrase, salt)
        a2 = argon2i(passphrase, salt)
        a3 = argon2i(passphrase, salt)
        assert a3 == a2
        assert (
            passphrase == b"\x00"
        )  # this should not pass but does (on first run only!) due to the pymonocypher bug where it is zeroing a (interned, due to being 1 byte) bytes object


        assert a1 == a2 # this should pass, but does not



import os
import unittest
from dataclasses import dataclass, asdict
from rendez.vous.reunion.primitives import Hash
from rendez.vous.reunion.session import ReunionSession, DEFAULT_HKDF_SALT


def DeterministicSession(
    passphrase: bytes, payload: bytes, seed: bytes, salt=DEFAULT_HKDF_SALT
):
    return ReunionSession(
        payload=payload,
        salt=salt,
        passphrase=passphrase,
        dh_seed=Hash(seed + b"dh"),
        ctidh_seed=Hash(seed + b"ct"),
        gamma_seed=Hash(seed + b"g"),
        delta_seed=Hash(seed + b"d"),
        dummy_seed=Hash(seed + b"d"),
        tweak=Hash(seed + b"t")[0],
    )


@dataclass
class SessionKAT:
    """
    This is a step toward we way that we could have regeneratable data-driven
    known-answer tests.

    Ideally the actual KATs could be read from a data file rather than this
    docstring, and could also include serialized secret keys (rather than just
    their seed as it is here now) so that other implementations could load them
    from the data file and reunion compatibility without necessarily needing to
    be completely seed-compatible when generating the keys.

    >>> SessionKAT.generate_from_seed(bytes.fromhex("1234"))
          seed: 1234
    passphrase: 9b0ac4fbd2e84a047b40695c391664890e570ee302a22c16c7025f52ed0586db
     payload_A: 00b7bf81a81bea5506669e6c00646beead875b9b1c5a8f30ea4a11ccf6ce2a98
     payload_B: 5c0d6c7b18d41aa1bb35dcf72f91bf9da292278c6057c8525df7e76ee4fa0764
           AT1: b7e50c69c1576ac0c38aa770facf18f86e837de3a265ba1b6776a4d64526c546214e6df3a72dad41d7218a65a691198ec1d13c040b100abbd8f72519bcfe597bc5b7012111e8b0e4bd59037650c24c047677ccccbe6f0230730ba7735edb73f03aee4c4115c4f0c9376d4ce59d311e90391fa65061df52d7ac6b4cd09b7e93a530a6fef8ddc407f67323ee079abe15669e9b15d9d4fec0e542ff69994234c609520e55fc3d434366794d59a0d672d7fd3dbad2eed6713de60de1098dc99bc2d24b24d1fdedb856c9104690350d0917d173de3ff6d4d5e6957305cd761eb4bdc2
           AT2: da4cc76fb741b2bb28ca95754290b1e484db06c9698254b7146d9ec36d5ad715
           AT3: f78bbb76c53025c3a0218aa54bbfedb02f3d8e375c737976272b1c0e4d885ec2
           BT1: 54ad5d6ca02c133782ffbbc26027d6d096ffb48c0df8c4bc469ba7b50881ea6bce6f8ba1e460862933a85b9b6a190adc6648cdd61a648e570d1b34b7e5830057c8117e95c57d7c1da083ad9d39ec2c367856b8dbe164f7f964c53da076f690edb1e8f4e276e757e074545eda23a60397227cf5d8d63d4ded364d01042402b9c0135cf19897d79ec764a37c579ebf3aaaee10926238dc06bc4d7b89b3b67cd7046b350f40fbb821fae6ec563366f419de62d3bdef94405a6f39595342fcb9fa6261185db123cd0e1d367097cf80f47f0acb9ba73062f605abc187556b65a9bc23
           BT2: 4eb24741bb5365e0424ce9010a43c06f00e12bfb326fc165efaed029294ba442
           BT3: 377a59cbbc4eeaa4ac25827b279dc115cb3b051449feb8248d5e2bc6f74460b6
    """

    seed: bytes = None
    passphrase: bytes = None
    payload_A: bytes = None
    payload_B: bytes = None
    AT1: bytes = None
    AT2: bytes = None
    AT3: bytes = None
    BT1: bytes = None
    BT2: bytes = None
    BT3: bytes = None

    @classmethod
    def generate_from_seed(cls, seed):
        self = cls()
        self.seed = seed
        self.passphrase = Hash(b"passphrase" + seed)
        self.payload_A = Hash(b"payload A" + seed)
        self.payload_B = Hash(b"payload B" + seed)
        A = DeterministicSession(self.passphrase, self.payload_A, b"a" + seed)
        B = DeterministicSession(self.passphrase, self.payload_B, b"b" + seed)
        self.AT1 = A.t1
        self.BT1 = B.t1
        self.AT2 = A.process_t1(B.t1)
        self.BT2 = B.process_t1(A.t1)
        self.AT3, a_isdummy = A.process_t2(B.t1.id, self.BT2)
        self.BT3, b_isdummy = B.process_t2(A.t1.id, self.AT2)
        assert not a_isdummy and not b_isdummy
        A.process_t3(B.t1.id, self.BT3)
        B.process_t3(A.t1.id, self.AT3)
        assert A.results[0] == self.payload_B
        assert B.results[0] == self.payload_A
        return self

    def __repr__(self):
        return "\n".join(f"{k:>10}: {bytes(v).hex()}" for k, v in asdict(self).items())



import os
import unittest

from rendez.vous.reunion.session import ReunionSession
from rendez.vous.reunion.util import get_pairs

from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase
from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase1
from rendez.vous.reunion.__vectors__ import ReunionSession_passphrase2
from rendez.vous.reunion.__vectors__ import ReunionSession_A_msg
from rendez.vous.reunion.__vectors__ import ReunionSession_B_msg
from rendez.vous.reunion.__vectors__ import ReunionSession_four_party_A_msg
from rendez.vous.reunion.__vectors__ import ReunionSession_four_party_B_msg
from rendez.vous.reunion.__vectors__ import ReunionSession_four_party_C_msg
from rendez.vous.reunion.__vectors__ import ReunionSession_four_party_D_msg

class TestReunionSession(unittest.TestCase):
    def setUp(self):
        pass

    def test_internals(self):
        A = ReunionSession.create(ReunionSession_passphrase, ReunionSession_A_msg)
        B = ReunionSession.create(ReunionSession_passphrase, ReunionSession_B_msg)

        A_t2 = A.process_t1(B.t1)
        B_t2 = B.process_t1(A.t1)

        A_t3, is_dummy_A = A.process_t2(B.t1.id, B_t2)
        B_t3, is_dummy_B = B.process_t2(A.t1.id, A_t2)

        assert is_dummy_A == False
        assert is_dummy_B == False

        A.process_t3(B.t1.id, B_t3)
        B.process_t3(A.t1.id, A_t3)

        A_peer = list(A.peers.values())[0]
        B_peer = list(B.peers.values())[0]
        assert A_peer.csidh_ss == B_peer.csidh_ss
        assert A_peer.dh_ss == B_peer.dh_ss
        assert B_peer.alpha_key == A.alpha_key
        assert B.salt == A.salt
        assert B.pdk == A.pdk

        A_msg_B = B.results[0]
        B_msg_A = A.results[0]

        self.assertEqual(ReunionSession_A_msg, A_msg_B)
        self.assertEqual(ReunionSession_B_msg, B_msg_A)

    def test_2party(self):
        """
        This test is disabled because the test_4party tests a superset of its
        functionality. It remains here as a demonstration of the simplest
        pairwise instantiation of the protocol.
        """
        A = ReunionSession.create(ReunionSession_passphrase, ReunionSession_A_msg)
        B = ReunionSession.create(ReunionSession_passphrase, ReunionSession_B_msg)

        A_t2 = A.process_t1(B.t1)
        B_t2 = B.process_t1(A.t1)

        A_t3, is_dummy_A = A.process_t2(B.t1.id, B_t2)
        B_t3, is_dummy_B = B.process_t2(A.t1.id, A_t2)

        A.process_t3(B.t1.id, B_t3)
        B.process_t3(A.t1.id, A_t3)

        A_msg_B = B.results[0]
        B_msg_A = A.results[0]

        self.assertEqual(ReunionSession_A_msg, A_msg_B)
        self.assertEqual(ReunionSession_B_msg, B_msg_A)

    def test_4party_interleaved(self):
        """
        4 parties means 16 CSIDH operations (N**2, or, N key generations plus
        N*(N-1) rendezvouses with others) so this test takes a little while.

        this test is often disabled as it is very similar to test_4party.

        in this variant, each pair of peers runs the protocol to completion
        before the next pair begins (but with each peer still reusing a single
        session).
        """

        A = ReunionSession.create(ReunionSession_passphrase1, ReunionSession_four_party_A_msg)
        B = ReunionSession.create(ReunionSession_passphrase1, ReunionSession_four_party_B_msg)
        C = ReunionSession.create(ReunionSession_passphrase2, ReunionSession_four_party_C_msg)
        D = ReunionSession.create(ReunionSession_passphrase2, ReunionSession_four_party_D_msg)

        sessions = (A, B, C, D)

        for a, b in get_pairs(sessions):
            a_t2 = a.process_t1(b.t1)
            b_t2 = b.process_t1(a.t1)

            a_t3, is_dummy_a = a.process_t2(b.t1.id, b_t2)
            b_t3, is_dummy_b = b.process_t2(a.t1.id, a_t2)

            a.process_t3(b.t1.id, b_t3)
            b.process_t3(a.t1.id, a_t3)

        A_msg_B = B.results[0]
        B_msg_A = A.results[0]
        C_msg_D = D.results[0]
        D_msg_C = C.results[0]

        self.assertEqual(ReunionSession_four_party_A_msg, A_msg_B)
        self.assertEqual(ReunionSession_four_party_B_msg, B_msg_A)
        self.assertEqual(ReunionSession_four_party_C_msg, C_msg_D)
        self.assertEqual(ReunionSession_four_party_D_msg, D_msg_C)
        self.assertTrue(all(len(r.results) == 1 for r in sessions))

    def test_4party(self):
        """
        4 parties means 16 CSIDH operations (N**2, or, N key generations plus
        N*(N-1) rendezvouses with others) so this test takes a little while.

        in this variant of the 4party test, we effectively operate in distinct
        phases: everyone transmits their t1 before anyone transmits their t2,
        etc.

        """

        # Phase 0: setup
        A = ReunionSession.create(ReunionSession_passphrase1, ReunionSession_four_party_A_msg)
        B = ReunionSession.create(ReunionSession_passphrase1, ReunionSession_four_party_B_msg)
        C = ReunionSession.create(ReunionSession_passphrase2, ReunionSession_four_party_C_msg)
        D = ReunionSession.create(ReunionSession_passphrase2, ReunionSession_four_party_D_msg)

        Rs = (A, B, C, D)

        # Phase 1: Transmit 𝑇1
        t1s = [r.t1 for r in Rs]

        # Phase 2: Process 𝑇1; transmit 𝑇2
        t2s = [
            (r.t1.id, t1.id, r.process_t1(t1)) for r in Rs for t1 in t1s if r.t1 != t1
        ]

        # Phase 3: Process 𝑇2, transmit 𝑇3
        t3s = [
            (r.t1.id, from_, r.process_t2(from_, t2)[0])
            for r in Rs
            for from_, to, t2 in t2s
            if r.t1.id == to
        ]

        # Phase 4: Process 𝑇3; decrypt payload
        [r.process_t3(from_, t3) for r in Rs for from_, to, t3 in t3s if r.t1.id == to]

        A_msg_B = B.results[0]
        B_msg_A = A.results[0]
        C_msg_D = D.results[0]
        D_msg_C = C.results[0]

        self.assertEqual(ReunionSession_four_party_A_msg, A_msg_B)
        self.assertEqual(ReunionSession_four_party_B_msg, B_msg_A)
        self.assertEqual(ReunionSession_four_party_C_msg, C_msg_D)
        self.assertEqual(ReunionSession_four_party_D_msg, D_msg_C)
        self.assertTrue(all(len(r.results) == 1 for r in Rs))


if __name__ == "__main__":
    unittest.main()



import os
import unittest

class TestDependencies(unittest.TestCase):
  def test_imports(self):
    import monocypher
    from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
    from flask import Flask, stream_with_context, request, Response
    from hashlib import blake2b, shake_256
    from highctidh import ctidh
    from hkdf import Hkdf
    from stem.control import Controller
    from typing import Dict, List
    from typing import Set
    import asyncio
    import click
    import datetime
    import highctidh
    import ifaddr
    import json
    import logging
    import monocypher
    import os
    import random
    import requests
    import secrets
    import socket
    import socks
    import struct
    import time

  def test_exports(self):
    from rendez.vous.reunion.primitives import aead_decrypt, aead_encrypt, argon2i
    from rendez.vous.reunion.primitives import generate_hidden_key_pair, Hash
    from rendez.vous.reunion.primitives import highctidh_deterministic_rng, hkdf, prp_decrypt
    from rendez.vous.reunion.primitives import prp_encrypt, unelligator, x25519




import os
import unittest

from rendez.vous.reunion.constants import DEFAULT_AEAD_NONCE, DEFAULT_ARGON_SALT, DEFAULT_HKDF_SALT
from rendez.vous.reunion.primitives import aead_decrypt, aead_encrypt, argon2i
from rendez.vous.reunion.primitives import generate_hidden_key_pair, Hash
from rendez.vous.reunion.primitives import highctidh_deterministic_rng, hkdf, prp_decrypt
from rendez.vous.reunion.primitives import prp_encrypt, unelligator, x25519
from rendez.vous.reunion.__vectors__ import esk_a_seed, esk_b_seed
from rendez.vous.reunion.__vectors__ import epk_a, epk_b
from rendez.vous.reunion.__vectors__ import esk_a, esk_b
from rendez.vous.reunion.__vectors__ import pk_a, pk_b
from rendez.vous.reunion.__vectors__ import aead_key, aead_pt, aead_ad, aead_ct
from rendez.vous.reunion.__vectors__ import prp_key, prp_pt, prp_ct
from rendez.vous.reunion.__vectors__ import a1, a2, pdk1, pdk2, h, h_preimage


class TestPrimitivesStaticVectors(unittest.TestCase):
    def test_hash(self):
        _hash = Hash(h_preimage)
        self.assertEqual(h, _hash)

    def test_aead_encrypt(self):
        _aead_ct = aead_encrypt(aead_key, aead_pt, aead_ad)
        self.assertEqual(aead_ct, _aead_ct)

    def test_aead_decrypt(self):
        _aead_ct = aead_encrypt(aead_key, aead_pt, aead_ad)
        _aead_pt = aead_decrypt(aead_key, aead_ct, aead_ad)
        self.assertEqual(aead_pt, _aead_pt)

    def test_elligator(self):
        # We must copy these because generate_hidden_key_pair will bzero the memory
        esk_a_seed_elligator_copy = bytes(a for a in esk_a_seed)
        esk_b_seed_elligator_copy = bytes(b for b in esk_b_seed)
        _epk_a, _esk_a = generate_hidden_key_pair(esk_a_seed_elligator_copy)
        _epk_b, _esk_b = generate_hidden_key_pair(esk_b_seed_elligator_copy)
        self.assertEqual(epk_a, _epk_a)
        self.assertEqual(esk_a, _esk_a)
        self.assertEqual(epk_b, _epk_b)
        self.assertEqual(esk_b, _esk_b)

    def test_unelligator(self):
        # We must copy these because generate_hidden_key_pair will bzero the memory
        esk_a_unelligator_seed_copy = bytes(a for a in esk_a_seed)
        esk_b_unelligator_seed_copy = bytes(b for b in esk_b_seed)
        _epk_a, _esk_a = generate_hidden_key_pair(esk_a_unelligator_seed_copy)
        _epk_b, _esk_b = generate_hidden_key_pair(esk_b_unelligator_seed_copy)
        self.assertEqual(epk_a, _epk_a)
        self.assertEqual(esk_a, _esk_a)
        self.assertEqual(epk_b, _epk_b)
        self.assertEqual(esk_b, _esk_b)
        self.assertEqual(pk_a, unelligator(_epk_a))
        self.assertEqual(pk_b, unelligator(_epk_b))

    def test_elligator_dh(self):
        # We must copy these because generate_hidden_key_pair will bzero the memory
        esk_a_seed_dh_copy = bytes(e for e in esk_a_seed)
        esk_b_seed_dh_copy = bytes(f for f in esk_b_seed)
        _epk_a, _esk_a = generate_hidden_key_pair(esk_a_seed_dh_copy)
        _epk_b, _esk_b = generate_hidden_key_pair(esk_b_seed_dh_copy)
        ss1 = x25519(esk_a, unelligator(epk_b))
        ss2 = x25519(esk_b, unelligator(epk_a))
        self.assertEqual(ss1, ss2)

    def test_prp(self):
        _ct = prp_encrypt(prp_key, prp_pt)
        self.assertEqual(prp_ct, _ct)
        _msg = prp_decrypt(prp_key, _ct)
        self.assertEqual(prp_pt, _msg)

    def test_argon_hkdf_with_internal_argon2i_bzero(self):
        salt_a1 = DEFAULT_HKDF_SALT
        salt_a2 = b'\x01' * 32
        passphrase_a1 = b'passphrase'
        passphrase_a2 = b'passphrase'
        _a1 = argon2i(passphrase_a1, salt_a1)
        self.assertEqual(a1, _a1)
        _a2 = argon2i(passphrase_a2, salt_a2)
        self.assertEqual(a2, _a2)
        passphrase_a1_copy = bytes(g for g in passphrase_a1)
        passphrase_a2_copy = bytes(h for h in passphrase_a1)
        _pdk1 = hkdf(argon2i(passphrase_a1_copy, salt_a1, _wipe=True), salt_a1).expand(b'', 32)
        _pdk2 = hkdf(argon2i(passphrase_a2_copy, salt_a1, _wipe=True), salt_a2).expand(b'', 32)
        self.assertEqual(passphrase_a1_copy, passphrase_a2_copy)
        self.assertEqual(pdk1, _pdk1)
        self.assertEqual(pdk2, _pdk2)

    def test_argon_hkdf_without_internal_argon2i_bzero(self):
        salt_a1 = DEFAULT_HKDF_SALT
        salt_a2 = b'\x01' * 32
        passphrase_a1 = b'passphrase'
        passphrase_a2 = b'passphrase'
        _a1 = argon2i(passphrase_a1, salt_a1)
        self.assertEqual(a1, _a1)
        _a2 = argon2i(passphrase_a2, salt_a2)
        self.assertEqual(a2, _a2)
        _pdk1 = hkdf(argon2i(passphrase_a1, salt_a1, _wipe=False), salt_a1).expand(b'', 32)
        _pdk2 = hkdf(argon2i(passphrase_a2, salt_a1, _wipe=False), salt_a2).expand(b'', 32)
        self.assertEqual(pdk1, _pdk1)
        self.assertEqual(pdk2, _pdk2)

    def test_argon_single_char_memory_corruption(self):
        passphrase = b'p'
        salt = b'\x00' * 32
        a1 = argon2i(passphrase, salt, _wipe=True)

        salt = b'\x00' * 32
        passphrase = b'p'
        # Here we do not zero memory
        a2 = argon2i(passphrase, salt, _wipe=False)
        self.assertEqual(passphrase, b'p')
        self.assertEqual(salt, b'\x00' * 32)

        # Here we do zero memory
        self.assertEqual(passphrase, b'p')
        a3 = argon2i(passphrase, salt, _wipe=True)
        # The salt value remains 32 bytes
        self.assertEqual(salt, b'\x00' * 32)
        # The password goes from length 32 to length 1
        self.assertEqual(passphrase, b'\x00')

        self.assertEqual(a3, a2)
        self.assertNotEqual(a1, a2)



"""
REUNION server

>>> assert logger is not None
"""
# TODO: Tor stuff is not finished yet.
# TODO: should probably persist messages to an on-disk thing to survive restarts

from rendez.vous.reunion.__version__ import __version__

from typing import Set

HASH_LEN = 32 # output of primitives.Hash()
MSG_LEN  = 96 + 64 # padded message size, would be nice if this could be given as a parameter to ReunionSession
KIND_CHUNK_SIZE = {
    't1': 32 + 64 + 16 + MSG_LEN + 16, # t1
    't2': HASH_LEN + 32, # Hash(t1) + t2
    't3': HASH_LEN + 32, # Hash(t1) + t3
}

import click
import asyncio
from flask import Flask, stream_with_context, request, Response

import logging
logger = logging.getLogger('reunion-server')

app = Flask(__name__)

t1s : Set[bytes] = set()
t2s : Set[bytes] = set()
t3s : Set[bytes] = set()

# @stream_with_context
# def stream_list(lst):
#   for item in lst:
#     yield lst

def status(ctx):
    global t1s, t2s, t3s
    logger.debug('%s: t1:%d t2:%d t3:%d', ctx, len(t1s), len(t2s), len(t3s))

def update(kind, items:bytes):
    '''Insert in sorted order, ignoring duplicates.'''
    status(kind.decode())
    if not items:
        # TODO actually this is fine: logger.error('no get_data() in update')
        return
    required_length = KIND_CHUNK_SIZE['t'+kind.decode()]
    if len(items) % required_length != 0:
        logger.error('length mismatch %d %% %d != 0',
                     len(items), required_length)
        return
    if b'1' == kind: lst = t1s
    elif b'2' == kind: lst = t2s
    elif b'3' == kind: lst = t3s
    else: assert False
    for item in (items[idx : idx + required_length]
                 for idx in range(0, len(items), required_length)):
        if item in lst:
            logger.debug('inserting new T%s: prefix %s',
                         kind.decode(),
                         item[:8].hex(), )
        lst.add(item)
    return True

@app.route('/t1', methods=['POST','GET'], endpoint='t1')
async def serve_t1():
    global t1s
    update(b'1', request.get_data())
    resp = Response(b''.join(t1s))
    resp.headers.set('Content-Type', 'application/octet-stream')
    return resp

@app.route('/t2', methods=['POST','GET'], endpoint='t2')
async def serve_t2():
    global t2s
    update(b'2', request.get_data())
    resp = Response(b''.join(t2s))
    resp.headers.set('Content-Type', 'application/octet-stream')
    return resp

@app.route('/t3', methods=['POST','GET'], endpoint='t3')
async def serve_t3():
    global t3s
    update(b'3', request.get_data())
    resp = Response(b''.join(t3s))
    resp.headers.set('Content-Type', 'application/octet-stream')
    return resp



def setup_tor(bind_ip, private_onion_key=None):
    logger.debug('[*] CONNECTING TO TOR CONTROLLER')
    from stem.control import Controller
    controller = Controller.from_port(address="127.0.0.1", port=9051)
    # after enabling ControlPort in torrc
    controller.authenticate(password="")

    # tor shared random (for epoch):
    # also available from
    # grep shared-rnd-(previous|current)-value /var/lib/tor/cached-microdesc-consensus

    # can use https://stem.torproject.org/api/descriptor/collector.html
    # to grab previous epoch microdescs:
    logger.debug('current rnd: %s', controller.get_info('sr/current'))
    # 'Ujtd/Zqjx0OlO4xubf9MBoqS12nraI6TYzyJwiZzsqA='
    logger.debug('previous rnd: %s', controller.get_info('sr/previous'))
    # 'mdr2FK+N8Pgxqy7kKqcNXWZIiHP1981LlHczEIwFkrc='

    # XXX: should use add_event_listener() to detect SRV changes

    hs_opts = {'key_type': 'NEW',
               'key_content': 'ED25519-V3',
               'detached': True,
               }
    if private_onion_key:
        hs_opts['key_type'] = 'ED25519-V3'
        hs_opts['key_content'] = private_onion_key

        # Remove all existing ephemeral hidden services (not ideal,
        # but required to avoid an exception when re-creating using
        # private key below):
        list(map(controller.remove_ephemeral_hidden_service,
                 controller.list_ephemeral_hidden_services(detached=True)))

    resp = controller.create_ephemeral_hidden_service(
        {1234: f'{bind_ip}:1234',},
        await_publication=False, **hs_opts)
    logger.debug('private key %s', resp.private_key or private_onion_key)
    hostname = resp.service_id + '.onion'
    logger.warn('listening on %s', hostname)

async def tcp_client_cb(reader, writer):
    global t1s
    global t2s
    global t3s
    print('got a connection', reader,'will now read kind')
    try:
        kind = await reader.readexactly(1)
    except asyncio.exceptions.IncompleteReadError:
        print('closing because bad')
        writer.close()
        return
    if kind not in (b'1', b'2', b'3'):
        print('wrong kind, closing')
        writer.close()
        return
    chunk_size = KIND_CHUNK_SIZE['t'+kind.decode()]
    # debug: kind
    status('tcp.t'+kind.decode())
    # TODO: update() looks at request.method etc, it shouldn't
    while reader.at_eof() != True:
        try:
            chunk = await reader.readexactly(chunk_size)
        except asyncio.exceptions.IncompleteReadError:
            break
        update(kind, chunk)
    if b'1' == kind:
        writer.write(b''.join(t1s))
    if b'2' == kind:
        writer.write(b''.join(t2s))
    if b'3' == kind:
        writer.write(b''.join(t3s))
    try: writer.write_eof()
    except: pass # that fails if they killed us
    await writer.drain()
    writer.close()
    await writer.wait_closed()
    logger.debug('goodbye to client %s %s', reader, writer)

async def serve_tcp(bind_ip):
    logger.debug('starting asyncio server')
    server = await asyncio.start_server(tcp_client_cb, host=bind_ip, port=1921,
                               reuse_address=True, reuse_port=True)
    while True:
        await asyncio.sleep(10)
    # async with server:
    #   await server.serve_forever()

def launch(mode, verbose, bind, private_onion_key):
    global logger
    if 'http' == mode:
        if verbose:
            app.debug=True
        logger = app.logger
        app.run()
    elif mode in ('tor', 'tcp'):
        if verbose:
            logger.setLevel(logging.DEBUG)
            logging.debug('for some reason this is needed to get logger to show debug/info wtf') # TODO
        if 'tor' == mode:
            setup_tor(bind, private_onion_key)
        asyncio.run(serve_tcp(bind))

@click.command()
@click.version_option(__version__)
@click.option("--bind", type=str, show_default=True,
              default='127.0.0.2',
              help="IP to bind to in TCP/tor mode", )
@click.option("--verbose", is_flag=True,
              help="Show verbose debug output", )
@click.option("--mode", type=str, default='http',
              show_default=True,
              help='http/tcp/tor', ) # TODO this should be a choice somehow
@click.option("--private-onion-key", type=str, default=None,
              help='key for the onion service address')
def server(**kw):
    """
    REUNION http/tcp/tor server
    """
    print('mode', kw['mode'])
    launch(**kw)

def main(**kw):
    server(**kw)

if '__main__' == __name__:
    import doctest
    doctest.testmod(verbose=True)



import struct
from hashlib import blake2b as _blake2b
from hashlib import shake_256 as _shake_256
from hkdf import Hkdf as _Hkdf
from highctidh import ctidh as _ctidh
import monocypher
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from rendez.vous.reunion.constants import DEFAULT_AEAD_NONCE, DEFAULT_ARGON_SALT
from rendez.vous.reunion.constants import DEFAULT_HKDF_SALT, DEFAULT_CTIDH_SIZE

ctidh1024 = _ctidh(DEFAULT_CTIDH_SIZE)

def Hash(msg: bytes) -> bytes:
    """
    *Hash* takes *msg* and returns 32 bytes of the *blake2b* digest of the
    message as bytes.

    note: Python's *blake2b* is blake2b 512 bit and we use only the last 256
    bits of the hash digest.

    >>> from rendez.vous.reunion.__vectors__ import h, h_preimage
    >>> _hash = Hash(h_preimage)
    >>> len(_hash) == 32
    True
    >>> type(_hash) == bytes
    True
    >>> h == _hash
    True
    """
    return _blake2b(msg).digest()[:32]

def argon2i(password: bytes, salt: bytes, _iterations: int = 3,
            _wipe: bool=False) -> bytes:
    """
    *argon2i* takes *password* of an arbitrary length encoded as bytes, a 32
    byte *salt*, and returns a 32 byte result encoded as bytes.

    REUNION does not negotiate the other parameters to argon2i.

    >>> from rendez.vous.reunion.__vectors__ import argon2i_password, argon2i_salt, argon2i_hash
    >>> _argon2i_hash = argon2i(argon2i_password, argon2i_salt)
    >>> argon2i_hash == _argon2i_hash
    True
    """
    return monocypher.argon2i_32(
        nb_blocks=100000,
        nb_iterations=_iterations,
        password=password,
        salt=salt,
        key=None,
        ad=None,
        _wipe=_wipe,
    )

def hkdf(key: bytes, salt: bytes, hash=_blake2b):
    """
    *hkdf* wraps a standard HKDF and uses *blake2b* by default.

    >>> from rendez.vous.reunion.__vectors__ import hkdf_salt, hkdf_key, hkdf_pdk, hkdf_pdk
    >>> _hkdf_result = hkdf(hkdf_key, hkdf_salt)
    >>> _hkdf_pdk = _hkdf_result.expand(b'', 32)
    >>> hkdf_pdk == _hkdf_pdk
    True
    """
    return _Hkdf(salt=salt, input_key_material=key, hash=hash)

def x25519(sk: bytes, pk: bytes) -> bytes:
    """
    *x25519* performs a Diffie-Hellman key-exchange between two parties that
    results in a 32 byte shared secret. The public key value *pk* should
    already be transformed from an elligator representation to a normal x25519
    public key with *unelligator*.

    >>> from rendez.vous.reunion.__vectors__ import x25519_sk_seed_a, x25519_sk_seed_b
    >>> epk_25519_a, sk_25519_a = generate_hidden_key_pair(x25519_sk_seed_a)
    >>> epk_25519_b, sk_25519_b = generate_hidden_key_pair(x25519_sk_seed_b)
    >>> pk_25519_a = unelligator(epk_25519_a)
    >>> pk_25519_b = unelligator(epk_25519_b)
    >>> shared_secret_a: bytes = x25519(sk_25519_a, pk_25519_b)
    >>> shared_secret_b: bytes = x25519(sk_25519_b, pk_25519_a)
    >>> shared_secret_a == shared_secret_b 
    True
    """
    return monocypher.key_exchange(sk, pk)

def aead_encrypt(key: bytes, plaintext: bytes, ad: bytes) -> bytes:
    """
    *aead_encrypt* takes *key*, *msg*, *ad* as bytes and returns *mac* and *ct*
    bytes objects.

    XChaCha20 and Poly1305 (RFC 8439)

    >>> from rendez.vous.reunion.__vectors__ import aead_ad, aead_key, aead_pt, aead_ct
    >>> _aead_ct = aead_encrypt(aead_key, aead_pt, aead_ad)
    >>> aead_ct == _aead_ct
    True
    """
    mac, ct = monocypher.lock(key, DEFAULT_AEAD_NONCE, plaintext, associated_data=ad)
    return mac + ct

def aead_decrypt(key: bytes, ciphertext: bytes, ad: bytes) -> bytes:
    """

    *aead_decrypt* takes *key*, *ciphertext*, *ad* as bytes and returns
    *plaintext* as bytes.

    XChaCha20 and Poly1305 (RFC 8439)

    >>> from rendez.vous.reunion.__vectors__ import aead_ad, aead_key, aead_pt, aead_ct
    >>> _aead_pt = aead_decrypt(aead_key, aead_ct, aead_ad)
    >>> aead_pt == _aead_pt
    True
    """
    mac, ct = ciphertext[:16], ciphertext[16:]
    return monocypher.unlock(key, DEFAULT_AEAD_NONCE, mac, ct, associated_data=ad)


def unelligator(hidden: bytes) -> bytes:
    """
    *unelligator* takes *hidden* a bytes object that contains a single x25519
    public key encoded with the elligator map; it reverses the map returning a
    bytes object that represents a normal x25519 public key.

    >>> from rendez.vous.reunion.__vectors__ import esk_a_seed, esk_b_seed
    >>> from rendez.vous.reunion.__vectors__ import epk_a, epk_b
    >>> from rendez.vous.reunion.__vectors__ import pk_a, pk_b
    >>> esk_a_seed_dt_copy = bytes(a for a in esk_a_seed)
    >>> esk_b_seed_dt_copy = bytes(b for b in esk_b_seed)
    >>> epk_25519_a, sk_25519_a = generate_hidden_key_pair(esk_a_seed_dt_copy)
    >>> epk_25519_a == epk_a
    True
    >>> epk_25519_b, sk_25519_b = generate_hidden_key_pair(esk_b_seed_dt_copy)
    >>> epk_25519_a == epk_a
    True
    >>> pk_25519_a = unelligator(epk_25519_a)
    >>> pk_25519_a == pk_a
    True
    >>> pk_25519_b = unelligator(epk_25519_b)
    >>> pk_25519_b == pk_b
    True
    """
    return monocypher.elligator_map(hidden)

def generate_hidden_key_pair(seed: bytes) -> bytes:
    """
    *generate_hidden_key_pair* takes a 32 byte object known as *seed* and
    returns a two-tuple consisting of a bytes object containing a x25519 public
    key encoded with the elligator map, and the corresponding bytes object for
    the respective x25519 secret key.

    >>> from rendez.vous.reunion.__vectors__ import hidden_key_pair_seed, hidden_key_pair_pk_a
    >>> from rendez.vous.reunion.__vectors__ import hidden_key_pair_sk_a
    >>> epk_25519_a, sk_25519_a = generate_hidden_key_pair(hidden_key_pair_seed)
    >>> hidden_key_pair_pk_a == epk_25519_a 
    True
    >>> hidden_key_pair_sk_a == sk_25519_a
    True
    """
    return monocypher.elligator_key_pair(seed)

def generate_ctidh_key_pair(seed: bytes) -> (object, object):
    """
    *generate_hidden_key_pair* takes a 32 byte object known as *seed* and
    returns a two-tuple consisting of a bytes object containing a CTIDH public
    key, and the corresponding bytes object for the respective CTIDH secret key.

    FIXME: it would be nice to upstream this function (and the CSPRNG it uses,
    defined later in this file) to the highctidh library.

    >>> from rendez.vous.reunion.__vectors__ import ctidh_key_pair_seed
    >>> from rendez.vous.reunion.__vectors__ import ctidh_key_pair_seed_pk
    >>> from rendez.vous.reunion.__vectors__ import ctidh_key_pair_seed_sk
    >>> pk, sk = generate_ctidh_key_pair(seed=ctidh_key_pair_seed)
    >>> ctidh_key_pair_seed_pk == bytes(pk)
    True
    >>> ctidh_key_pair_seed_sk == bytes(sk)
    True
    """
    rng = highctidh_deterministic_rng(seed)
    sk = ctidh1024.generate_secret_key(rng=rng, context=1)
    pk = ctidh1024.derive_public_key(sk)
    return pk, sk


def prp_encrypt(key: bytes, plaintext: bytes):
    """
    *prp_encrypt* takes *key* and *plaintext* and returns a bytes encoded
    ciphertext of length 32.  It is explictly not authenticated encryption by
    design and should only be used where authentication of ciphertexts is an
    anti-feature.

    This is intended to be implemented with rijndael with a 256 bit block size.
    Currently we use two 128 bit blocks of AES as rijndael is not in the
    standard library; it should be replaced with rijndael.

    >>> from rendez.vous.reunion.__vectors__ import prp_key, prp_pt, prp_ct
    >>> _prp_ct = prp_encrypt(prp_key, prp_pt)
    >>> prp_ct == _prp_ct
    True
    """
    assert len(key) == 32, len(key)
    assert len(plaintext) == 32, len(plaintext)
    cipher = Cipher(algorithms.AES(key), modes.ECB())
    encryptor = cipher.encryptor()
    res = encryptor.update(plaintext) + encryptor.finalize()
    assert len(res) == 32, len(res)
    return res

def prp_decrypt(key: bytes, ciphertext: bytes):
    """
    *prp_decrypt* takes *key* and *ciphertext* and returns a bytes encoded
    plaintext of length 32.  It is explictly not authenticated encryption by
    design and should only be used where authentication of ciphertexts is an
    anti-feature.

    This is intended to be implemented with rijndael with a 256 bit block size.
    Currently we use two 128 bit blocks of AES as rijndael is not in the
    standard library; it should be replaced with rijndael.

    >>> from rendez.vous.reunion.__vectors__ import prp_key, prp_pt, prp_ct
    >>> _prp_ct = prp_encrypt(prp_key, prp_pt)
    >>> prp_ct == _prp_ct
    True
    >>> _prp_pt = prp_decrypt(prp_key, prp_ct)
    >>> prp_pt == _prp_pt
    True
    """
    assert len(ciphertext) == 32, len(ciphertext)
    assert len(key) == 32, len(key)
    cipher = Cipher(algorithms.AES(key), modes.ECB())
    decryptor = cipher.decryptor()
    return decryptor.update(ciphertext) + decryptor.finalize()

def highctidh_deterministic_rng(seed: bytes):
    """
    *highctidh_deterministic_rng* takes a *seed* of at least 32 bytes and
    returns a generator suitable for deterministic outputs.

    This function was copied from a file in the examples directory in a branch
    of the public domain highctidh repo, and is used only to enable
    known-answer tests.

    Instantiate a SHAKE-256-based CSPRNG using a seed.  The seed should be at
    least 32 bytes (256 bits).

    Returns a function suitable for the optional rng= argument to
    highctidh.ctidh.generate_secret_key.  This enables deterministic key
    generation when also passing a deterministic context= argument.

    The CSPRNG keeps state internally to be able to provide unique entropy to
    libhighctidh (which calls it many times during the process of generating a
    key).

    It is safe to use the same seed to generate multiple keys if (and only if)
    **distinct** context arguments are passed.

    >>> from rendez.vous.reunion.__vectors__ import highctidh_context, highctidh_drng_seed
    >>> det_rng = highctidh_deterministic_rng(highctidh_drng_seed)
    >>> highctidh_1024_priv_key = ctidh1024.generate_secret_key(rng=det_rng, context=highctidh_context)
    """
    assert len(seed) >= 32, "deterministic seed should be at least 256 bits"
    context_state = {}
    def _shake256_csprng(buf: memoryview, context: int):
        """
        *_shake256_csprng* takes a memoryview *buf* and an integer in *context* and returns
        a function suitable for use with the highctidh determininstic *rng*
        parameter.

        >>> from rendez.vous.reunion.__vectors__ import highctidh_context, highctidh_drng_seed
        >>> det_rng = highctidh_deterministic_rng(highctidh_drng_seed)
        >>> highctidh_1024_priv_key = ctidh1024.generate_secret_key(rng=det_rng, context=highctidh_context) 
        """
        # context_state[context] is a counter, incremented on each call,
        # packed to little-endian uint64
        context_state[context] = 1 + context_state.get(context, 0)
        portable_state = struct.pack('<Q', context_state[context])
        # the user provided context packed to little-endian uint64:
        portable_context = struct.pack('<Q', context)
        little_endian_out = _shake_256(
            portable_context + portable_state + seed
        ).digest(len(buf))
        # interpret as little-endian uint32 tuples
        # and pack to native byte order as expected by libhighctidh.
        # This is required to get deterministic keys independent of the
        # endian-ness of the host machine:
        for i in range(0, len(buf), 4):
            portable_uint32 = struct.unpack('<L',little_endian_out[i:i+4])[0]
            buf[i:i+4] = struct.pack(
                '=L', portable_uint32)
    return _shake256_csprng



"""
REUNION client

>>> assert logger is not None
"""

# TODO: do something about hardcoded message length (96)
# TODO: more click options for ports, ips, etc.
# TODO: add epoch handling.
# TODO: port the multicast client over here (send_udp_multicast?)

from rendez.vous import reunion
from rendez.vous.reunion.primitives import Hash

# TODO: Hash should have a property specifying byte length
from rendez.vous.reunion.session import ReunionSession, T1
from rendez.vous.reunion.__version__ import __version__

# process_t1: should probably case to T1()
import requests
import time
import random
import asyncio

import datetime
import json  # for the output
import click # for the cli
import socks

import logging
logger = logging.getLogger('reunion-client')



def through_socks_proxy(socks_addr, socks_port, target_addr, target_port,
                        username=None, password=None):
    s = socks.socksocket()
    # set_proxy(proxy_type, addr[, port[, rdns[, username[, password]]]])
    s.set_proxy(socks.SOCKS4, socks_addr, socks_port, rdns=True,
                username=username, password=password)
    print((target_addr, target_port))
    s.connect((target_addr, target_port))
    return asyncio.open_connection(socks=s)

async def send_tor(where, what, chunk_size):
    rw = await through_socks_proxy(
        '127.0.0.1', 9050,
        '62fkbxb2jqxnjvigkpihj32qoyv4ht576im3vi7oq7zbyf2tg7kwopyd.onion', 1921,
        username='reunion',)
    logger.debug('have a socks connection %s', rw)
    async for x in send_tcp(where, what, chunk_size,
                            reader_writer=rw):
        logger.debug('yielded from send_tcp %s', x)
        yield x

async def send_tcp(where, what, chunk_size, reader_writer=None):
    '''chunk_size is used to group the results'''
    async def internal():
        if not reader_writer:
            reader_writer = await asyncio.open_connection('127.0.0.2', '1921')
        else:
            reader, writer = reader_writer
        logger.debug('got a connection %s', reader)
        writer.write(where[1].encode())
        for record in what:
            writer.write(record)
        writer.write_eof()
        await writer.drain()
        logger.debug('wrote records')
        while not reader.at_eof():
            logger.debug('reading more')
            try:
                record = await reader.readexactly(chunk_size)
            except asyncio.exceptions.IncompleteReadError:
                logger.debug('incomplete')
                break # EOF I guess?
            logger.debug('yielding record')
            yield record
        writer.close()
    # we ignore connection errors since we'll eventually converge.
    try:
        async for x in internal():
            yield x
    except ConnectionRefusedError as e:
        logger.warning('connection refused %s', e)
    except ConnectionResetError:
        logger.warning('connection reset during %s', where)
    logger.debug('done with send %s', where)

async def send_http(where, what, chunk_size):
    '''chunk_size is used to group the results'''
    assert len(set(what)) == len(what)
    try:
        p = requests.post('http://127.0.0.1:5000/'+where,
                          data=b''.join(what))
    except Exception as e:
        # TODO log that it failed
        logger.warning('request failed', e)
        return
    if not p.ok:
        return
    for idx in range(0, len(p.content), chunk_size):
        yield p.content[idx : idx + chunk_size]

async def store_result(result:bytes):
    try:
        dumpable = ['utf8', result.decode()]
    except UnicodeDecodeError as e:
        logger.warn('result decode error: %s', e)
        # fallback to hex in case we can't decode it as unicode
        dumpable = ['hex', result.hex()]
    print(json.dumps(dumpable))

async def launch(passphrase, message, mode, verbose, duration:int,
                 mandatory_sleep=0.5, randomized_sleep=2):
    start_time = datetime.datetime.now()
    duration = datetime.timedelta(seconds=duration)

    if verbose:
        logger.setLevel(logging.DEBUG)
        logging.debug('TODO why is this needed to get logger working?')
    logger.info(f'Running Reunion in {mode} mode for {duration}')
    if 'http' == mode:
        send = send_http
    elif 'tcp' == mode:
        send = send_tcp
    elif 'tor' == mode:
        send = send_tor
    else:
        raise Exception(f'invalid --mode {repr(mode)}')

    Me = ReunionSession.create(passphrase.encode(),
                               message.encode()[:96].ljust(96) # TODO truncation?
                               # TODO pass in salt= epoch id here, e.g. tor shared random
                               )
    my_t1 = Me.t1 # Me.t1.ljust(256)
    own_t1_hash = my_t1.id
    logger.debug('my_t1: %s', my_t1[:8].hex())

    answered_t1s = set( (own_t1_hash,) ) # don't answer own T1
    answered_t2s = set( )
    answered_t3s = set( )

    while datetime.datetime.now() < start_time + duration:
        t2s = []
        logger.debug('t1s, my_t1 is %s', my_t1[:8].hex())
        stat_current_t1s = len(answered_t1s)
        async for t1 in send('t1', [my_t1], len(my_t1)):
            t1 = T1(t1)
            # TODO assert own_t1_hash in response
            logger.debug('Seen T1 %s', t1[:8].hex())
            if t1.id in answered_t1s:
                continue # skip already processed
            try:
                t2 = Me.process_t1(t1)
            except:
                # process_t1 can throw:
                #   File "reunion/session.py", line 250, in __init__
                #    csidh_ss = csidh.dh(session.csidh_sk, csidh_pk)
                #  File "sibc/sibc/csidh/__init__.py", line 118, in dh
                #    ss = self.curve.coeff(self.gae.GAE_at_A(sk, pk)).x.to_bytes(
                #  File "sibc/sibc/csidh/gae_df.py", line 64, in GAE_at_A
                #    assert self.curve.issupersingular(A), "non-supersingular input curve"
                # AssertionError: non-supersingular input curve
                continue # skip to next T1
            answered_t1s.add(t1.id) # TODO shouldn't update before we manage to send
            answered_t2s.add(own_t1_hash + t2) # we'll skip T2s we emitted
            t2s.append(own_t1_hash + t2)
        logger.debug('answered %d/%d T1s, moving on to T2',
                     len(answered_t1s) - stat_current_t1s,
                     len(answered_t1s))
        t3s = []
        async for t2 in send('t2', t2s, 64):
            t2_id = t2[:32]
            if t2 in answered_t2s:
                continue # skipping already processed
            t3, is_dummy = Me.process_t2(t2_id, t2[32:])
            t3s.append(own_t1_hash + t3)
            answered_t3s.add(own_t1_hash + t3) # we'll skip our own uploads
            answered_t2s.add(t2)
        logger.debug('answered %d/%d T2s, moving on to T3', len(t3s), len(answered_t3s))
        # TODO this is where we should update answered_t2s
        stat_current_t3s = len(answered_t3s)
        async for t3 in send('t3', t3s, 64):
            t3_id, t3_payload = t3[:32], t3[32:]
            if t3 in answered_t3s:
                continue # skipping already processed
            processed_result = Me.process_t3(t3_id, t3_payload)
            logger.debug('t3: %s %s', t3_id.hex(), processed_result)
            if processed_result:
                await store_result(processed_result)
            answered_t3s.add(t3)
        logger.debug('processed %d/%d T3s, looping',
                     len(answered_t3s) - stat_current_t3s,
                     len(answered_t3s))
        # TODO should update answered_t3s here
        await asyncio.sleep(mandatory_sleep + randomized_sleep*random.random())
        logger.debug('looping')

@click.command()
@click.version_option(__version__)
@click.option("--duration", type=int,
              default=24*3600,
              show_default=True,
              help="Duration to run the protocol for", )
@click.option("--verbose", is_flag=True,
              help="Show verbose debug output", )
@click.option("--mode", type=str, default='http',
              show_default=True,
              help='http/tcp/tor', )
@click.option("--message", prompt=True, required=True, type=str,
              help="Message to deliver to people using the same password", )
@click.option("--passphrase", prompt=True, required=True, type=str, help="Passphrase/password")
def client(**kw):
    """
    REUNION http/tcp/tor client
    """
    asyncio.run(launch(**kw))

def main(**kw):
    client(**kw)

if '__main__' == __name__:
    import doctest
    doctest.testmod(verbose=True)
